"""Module for handling the GOG service"""

import json
import os
import time
import typing
from collections import defaultdict
from gettext import gettext as _
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import parse_qsl, unquote, urlencode, urlparse

from lxml import etree

from lutris import settings
from lutris.exceptions import AuthenticationError, UnavailableGameError
from lutris.installer import AUTO_ELF_EXE, AUTO_WIN32_EXE
from lutris.installer.installer_file import InstallerFile
from lutris.installer.installer_file_collection import InstallerFileCollection
from lutris.runners import get_runner_human_name
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import i18n, system
from lutris.util.http import HTTPError, Request, UnauthorizedAccessError
from lutris.util.log import logger
from lutris.util.strings import human_size, slugify

if typing.TYPE_CHECKING:
    from lutris.installer.installer import LutrisInstaller


class GogSmallBanner(ServiceMedia):
    """Small size game logo"""

    service = "gog"
    size = (100, 60)
    dest_path = os.path.join(settings.CACHE_DIR, "gog/banners/small")
    file_patterns = ["%s.jpg"]
    api_field = "image"
    url_pattern = "https:%s_prof_game_100x60.jpg"


class GogMediumBanner(GogSmallBanner):
    """Medium size game logo"""

    size = (196, 110)
    dest_path = os.path.join(settings.CACHE_DIR, "gog/banners/medium")
    url_pattern = "https:%s_196.jpg"


class GogLargeBanner(GogSmallBanner):
    """Big size game logo"""

    size = (392, 220)
    dest_path = os.path.join(settings.CACHE_DIR, "gog/banners/large")
    url_pattern = "https:%s_392.jpg"


class GOGGame(ServiceGame):
    """Representation of a GOG game"""

    service = "gog"

    @classmethod
    def new_from_gog_game(cls, gog_game):
        """Return a GOG game instance from the API info"""
        service_game = GOGGame()
        service_game.appid = str(gog_game["id"])
        service_game.slug = gog_game["slug"]
        service_game.name = gog_game["title"]
        service_game.details = json.dumps(gog_game)
        return service_game


class GOGService(OnlineService):
    """Service class for GOG"""

    id = "gog"
    name = _("GOG")
    icon = "gog"
    has_extras = True
    drm_free = True
    medias = {"banner_small": GogSmallBanner, "banner": GogMediumBanner, "banner_large": GogLargeBanner}
    default_format = "banner"

    embed_url = "https://embed.gog.com"
    api_url = "https://api.gog.com"

    client_id = "46899977096215655"
    client_secret = "9d85c43b1482497dbbce61f6e4aa173a433796eeae2ca8c5f6129f2dc4de46d9"
    redirect_uris = ["https://embed.gog.com/on_login_success?origin=client"]

    login_success_url = "https://www.gog.com/on_login_success"
    cookies_path = os.path.join(settings.CACHE_DIR, ".gog.auth")
    token_path = os.path.join(settings.CACHE_DIR, ".gog.token")
    cache_path = os.path.join(settings.CACHE_DIR, "gog-library.json")

    runner_to_os_dict = {"wine": "windows", "linux": "linux"}

    def __init__(self):
        super().__init__()

        gog_locales = {
            "en": "en-US",
            "de": "de-DE",
            "fr": "fr-FR",
            "pl": "pl-PL",
            "ru": "ru-RU",
            "zh": "zh-Hans",
        }
        self.locale = gog_locales.get(i18n.get_lang(), "en-US")

    @property
    def login_url(self):
        """Return authentication URL"""
        params = {
            "client_id": self.client_id,
            "layout": "client2",
            "redirect_uri": self.redirect_uris[0],
            "response_type": "code",
        }
        return "https://auth.gog.com/auth?" + urlencode(params)

    @property
    def credential_files(self) -> List[str]:
        return [self.cookies_path, self.token_path]

    def is_connected(self) -> bool:
        """Return whether the user is authenticated and if the service is available"""
        if not self.is_authenticated():
            return False
        try:
            user_data = self.get_user_data()
        except AuthenticationError:
            return False
        except UnauthorizedAccessError:
            logger.warning("GOG token is invalid")
            return False
        except HTTPError as ex:
            logger.exception("Failed to query user data: %s", ex)
            return False
        return bool(user_data and "username" in user_data)

    def load(self) -> List[GOGGame]:
        """Load the user game library from the GOG API"""
        if not self.is_connected():
            logger.error("User not connected to GOG")
            return []
        games = [GOGGame.new_from_gog_game(game) for game in self.get_library()]
        for game in games:
            game.save()
        self.match_games()
        return games

    def login_callback(self, url) -> None:
        return self.request_token(url)

    def request_token(self, url: str = "", refresh_token: str = "") -> None:
        """Get authentication token from GOG"""
        if refresh_token:
            grant_type = "refresh_token"
            extra_params = {"refresh_token": refresh_token}
        else:
            grant_type = "authorization_code"
            parsed_url = urlparse(url)
            response_params = dict(parse_qsl(parsed_url.query))
            if "code" not in response_params:
                logger.error("code not received from GOG")
                logger.error(response_params)
                return
            extra_params = {
                "code": response_params["code"],
                "redirect_uri": self.redirect_uris[0],
            }

        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": grant_type,
        }
        params.update(extra_params)
        url = "https://auth.gog.com/token?" + urlencode(params)
        request = Request(url, redacted_query_parameters=("refresh_token", "code"))
        try:
            request.get()
        except HTTPError as http_error:
            logger.error(http_error)
            logger.error("Failed to get token.")
            logger.warning("Clearing existing credentials")
            self.logout()
            return

        token = request.json
        with open(self.token_path, "w", encoding="utf-8") as token_file:
            token_file.write(json.dumps(token))
        if not refresh_token:
            SERVICE_LOGIN.fire(self)

    def load_token(self) -> dict:
        """Load token from disk"""
        if not os.path.exists(self.token_path):
            raise AuthenticationError("No GOG token available")

        with open(self.token_path, encoding="utf-8") as token_file:
            token_content = json.loads(token_file.read())

        if not token_content:
            raise AuthenticationError("No GOG token available")

        return token_content

    def get_token_age(self) -> float:
        """Return age of token"""
        token_stat = os.stat(self.token_path)
        token_modified = token_stat.st_mtime
        return time.time() - token_modified

    def make_request(self, url: str) -> Any:
        """Send a cookie authenticated HTTP request to GOG"""
        request = Request(url, cookies=self.load_cookies())
        request.get()
        if request.content.startswith(b"<"):
            raise AuthenticationError("Token expired, please log in again")
        return request.json

    def make_api_request(self, url: str) -> Any:
        """Send a token authenticated request to GOG"""
        token = self.load_token()

        if self.get_token_age() > 2600:
            self.request_token(refresh_token=token["refresh_token"])
            token = self.load_token()
        headers = {"Authorization": "Bearer " + token["access_token"]}
        request = Request(url, headers=headers, cookies=self.load_cookies())
        request.get()
        return request.json

    def get_user_data(self) -> dict:
        """Return GOG profile information"""
        url = "https://embed.gog.com/userData.json"
        return self.make_api_request(url)

    def get_library(self) -> List[dict]:
        """Return the user's library of GOG games"""
        if system.path_exists(self.cache_path):
            logger.debug("Returning cached GOG library")
            with open(self.cache_path, "r", encoding="utf-8") as gog_cache:
                return json.load(gog_cache)

        total_pages = 1
        games = []
        page = 1
        while page <= total_pages:
            products_response = self.get_products_page(page=page)
            page += 1
            total_pages = products_response["totalPages"]
            games += products_response["products"]
        with open(self.cache_path, "w", encoding="utf-8") as gog_cache:
            json.dump(games, gog_cache)
        return games

    def get_service_game(self, gog_game: dict) -> GOGGame:
        return GOGGame.new_from_gog_game(gog_game)

    def get_products_page(self, page: int = 1, search: str = None) -> Any:
        """Return a single page of games"""
        if not self.is_authenticated():
            raise AuthenticationError("User is not logged in")
        params = {"mediaType": "1"}
        if page:
            params["page"] = page
        if search:
            params["search"] = search
        url = self.embed_url + "/account/getFilteredProducts?" + urlencode(params)
        return self.make_request(url)

    def get_game_dlcs(self, product_id: str) -> List[dict]:
        """Return the list of DLC products for a game"""
        game_details = self.get_game_details(product_id)
        if not game_details["dlcs"]:
            return []
        all_products_url = game_details["dlcs"]["expanded_all_products_url"]
        return self.make_api_request(all_products_url)

    def get_game_details(self, product_id: str) -> dict:
        """Return game information for a given game"""
        if not product_id:
            raise ValueError("Missing product ID")
        logger.info("Getting game details for %s", product_id)
        url = "{}/products/{}?expand=downloads&locale={}".format(self.api_url, product_id, self.locale)
        return self.make_api_request(url)

    def get_download_info(self, downlink: str) -> List[dict]:
        """Return file download information, a list of dict containing the 'url' and
        'filename' for each file."""
        logger.info("Getting download info for %s", downlink)
        try:
            response = self.make_api_request(downlink)
        except HTTPError as ex:
            logger.error("HTTP error: %s", ex)
            return []
        if not response:
            logger.error("No download info obtained for %s", downlink)
            return []

        expanded = []
        for field in ("checksum", "downlink"):
            field_url = response[field]
            parsed = urlparse(field_url)
            query = dict(parse_qsl(parsed.query))
            exp = {"url": response[field]}

            if "path" in query:
                exp["filename"] = os.path.basename(query["path"])
            else:
                quoted_name = os.path.basename(parsed.path)
                filename = unquote(quoted_name)
                exp["filename"] = filename
                # Prior releases of Lutris did not unquote the name, so we'll provide the unquoted
                # name as an alternate so we can use download files that have that name, and
                # not re-download them.
                exp["alternate_filenames"] = [quoted_name] if quoted_name != filename else []

            expanded.append(exp)
        return expanded

    def get_downloads(self, gogid: str) -> dict:
        """Return all available downloads for a GOG ID"""
        if not gogid:
            logger.warning("Unable to get GOG data because no GOG ID is available")
            return {}
        gog_data = self.get_game_details(gogid)
        if not gog_data:
            logger.warning("Unable to get GOG data for game %s", gogid)
            return {}
        return gog_data["downloads"]

    def get_extras(self, gogid: str) -> Dict[str, List[dict]]:
        """Return a list of bonus content available for a GOG ID and its DLCs"""
        logger.debug("Download extras for GOG ID %s and its DLCs", gogid)
        game = self.get_game_details(gogid)
        if not game:
            logger.warning("Unable to get GOG data for game %s", gogid)
            return {}

        try:
            dlcs = self.get_game_dlcs(gogid)
        except Exception as ex:
            # I would rather just propagate this, but I'm fixing code from
            # commit f7842ec6 which says "don't crash when user isn't logged
            # in to GOG" so I guess we'll try to provide what extras we can,
            # without DLCs.
            logger.exception("Unable to retrieve DLC: %s", ex)
            dlcs = {}

        products = [game, *dlcs] if dlcs else [game]
        all_extras = {}
        for product in products:
            # Extras for DLCs you don't own are listed, but are not installable.
            if product.get("is_installable"):
                extras = [
                    {
                        "name": download.get("name", "").strip().capitalize(),
                        "type": download.get("type", "").strip(),
                        "total_size": download.get("total_size", 0),
                        "id": str(download["id"]),
                        "downlinks": [f.get("downlink") for f in download.get("files") or []],
                    }
                    for download in product["downloads"].get("bonus_content") or []
                ]
                if extras:
                    all_extras[product.get("title", "").strip()] = extras
        return all_extras

    def get_installers(self, downloads: Dict[str, List[dict]], runner: str, language: str = "en") -> List[dict]:
        """Return available installers for a GOG game"""
        # Filter out Mac installers
        gog_installers = [installer for installer in downloads.get("installers", []) if installer["os"] != "mac"]
        # Prefer native Linux installers if available, otherwise fall back to Windows
        # (scripts may need Windows installers for data extraction, engine reimplementations, etc.)
        linux_installers = [installer for installer in gog_installers if installer["os"] == "linux"]
        if linux_installers:
            gog_installers = linux_installers
        return [
            installer
            for installer in gog_installers
            if installer["language"] == self.determine_language_installer(gog_installers, language)
        ]

    def get_update_versions(self, gog_id: str, runner_name: Optional[str]) -> Dict[str, list]:
        """Return updates available for a game, keyed by patch version"""

        filter_os = self.runner_to_os_dict.get(runner_name) if runner_name else None

        games_detail = self.get_game_details(gog_id)
        patches = games_detail["downloads"]["patches"]
        if not patches:
            logger.info("No patches for %s", games_detail)
            return {}
        patch_versions = defaultdict(list)
        for patch in patches:
            if filter_os:
                patch_os = patch.get("os")
                if patch_os and filter_os != patch_os.casefold():
                    continue
            patch_versions[patch["name"]].append(patch)
        return patch_versions

    def determine_language_installer(self, gog_installers: List[dict], default_language: str = "en") -> str:
        """Return locale language string if available in gog_installers"""
        language = i18n.get_lang()
        gog_installers = [installer for installer in gog_installers if installer["language"] == language]
        if not gog_installers:
            language = default_language
        return language

    def query_download_links(self, download: Dict[str, List[dict]]) -> List[dict]:
        """Convert files from the GOG API to a format compatible with lutris installers"""
        download_links = []
        for game_file in download.get("files", []):
            downlink = game_file.get("downlink")
            if not downlink:
                logger.error("No download information for %s", game_file)
                continue
            for info in self.get_download_info(downlink):
                download_links.append(
                    {
                        "name": download.get("name", ""),
                        "os": download.get("os", ""),
                        "type": download.get("type", ""),
                        "total_size": download.get("total_size", 0),
                        "id": str(game_file["id"]),
                        "url": info["url"],
                        "filename": info["filename"],
                        "alternate_filenames": info["alternate_filenames"],
                    }
                )
        return download_links

    def get_extra_files(self, installer: "LutrisInstaller", selected_extras: List[dict]) -> List[InstallerFile]:
        extra_files = []
        for extra in selected_extras:
            downlinks = extra.get("downlinks")
            if downlinks:
                links = [info for link in downlinks for info in self.get_download_info(link)]
            elif str(extra["id"]) in selected_extras:
                links = self.query_download_links(extra)
            else:
                links = []

            if not links:
                logger.error("No download link for bonus content '%s' could be obtained.", extra.get("name"))

            for link in links:
                if link["filename"].endswith(".xml"):
                    # GOG gives a link for checksum XML files for bonus content
                    # but downloading them results in a 404 error.
                    continue
                extra_files.append(
                    InstallerFile(
                        installer.game_slug,
                        str(extra["id"]),
                        {
                            "url": link["url"],
                            "filename": link["filename"],
                            "alternate_filenames": link["alternate_filenames"],
                        },
                    )
                )
        return extra_files

    def _get_installer_links(self, installer: "LutrisInstaller", downloads: dict) -> List[dict]:
        """Return links to downloadable files from a list of downloads"""
        try:
            gog_installers = self.get_installers(downloads, installer.runner)
            if not gog_installers:
                return []
            if len(gog_installers) > 1:
                logger.warning("More than 1 GOG installer found, picking first.")
            _installer = gog_installers[0]
            return self.query_download_links(_installer)
        except HTTPError as err:
            raise UnavailableGameError(_("Couldn't load the download links for this game")) from err

    def get_patch_files(self, installer: "LutrisInstaller", installer_file_id: str) -> List[InstallerFile]:
        logger.debug("Getting patches for %s", installer.version)
        downloads = self.get_downloads(installer.service_appid)
        links = []
        for patch_file in downloads["patches"]:
            if "GOG " + patch_file["version"] == installer.version:
                links += self.query_download_links(patch_file)
        return self._format_links(installer, installer_file_id, links)

    def _format_links(
        self, installer: "LutrisInstaller", installer_file_id: str, links: List[dict]
    ) -> List[InstallerFile]:
        _installer_files = defaultdict(dict)  # keyed by filename
        for link in links:
            try:
                filename = link["filename"]
            except KeyError:
                logger.error("Invalid link: %s", link)
                raise
            if filename.lower().endswith(".xml"):
                if filename != installer_file_id:
                    filename = filename[:-4]
                _installer_files[filename]["checksum_url"] = link["url"]
                continue
            _installer_files[filename]["id"] = link["id"]
            _installer_files[filename]["url"] = link["url"]
            _installer_files[filename]["filename"] = filename
            _installer_files[filename]["alternate_filenames"] = link["alternate_filenames"]
            _installer_files[filename]["total_size"] = link["total_size"]
        files = []
        file_id_provided = False  # Only assign installer_file_id once
        for _file_id in _installer_files:
            installer_file = _installer_files[_file_id]
            if "url" not in installer_file:
                raise ValueError("Invalid installer file %s" % installer_file)
            filename = installer_file["filename"]
            if filename.lower().endswith((".exe", ".sh")) and not file_id_provided:
                file_id = installer_file_id
                file_id_provided = True
            else:
                file_id = _file_id
            files.append(
                InstallerFile(
                    installer.game_slug,
                    file_id,
                    {
                        "url": installer_file["url"],
                        "filename": installer_file["filename"],
                        "alternate_filenames": installer_file["alternate_filenames"],
                        "checksum_url": installer_file.get("checksum_url"),
                        "total_size": installer_file["total_size"],
                    },
                )
            )
        if not file_id_provided:
            raise UnavailableGameError(_("Unable to determine correct file to launch installer"))
        return files

    def get_installer_files(
        self, installer: "LutrisInstaller", installer_file_id: str, selected_extras: List[dict]
    ) -> Tuple[List[InstallerFileCollection], List[InstallerFile]]:
        try:
            downloads = self.get_downloads(installer.service_appid)
        except HTTPError as err:
            raise UnavailableGameError(_("Couldn't load the downloads for this game")) from err
        links = self._get_installer_links(installer, downloads)
        if links:
            formatted = self._format_links(installer, installer_file_id, links)
            files = [InstallerFileCollection(installer.game_slug, installer_file_id, formatted)]
        else:
            files = []

        extra_files = []
        if selected_extras:
            for extra_file in self.get_extra_files(installer, selected_extras):
                extra_files.append(extra_file)

        return files, extra_files

    def read_file_checksum(self, file_path: str) -> Tuple[str, str]:
        """Return the MD5 checksum for a GOG file
        Requires a GOG XML file as input
        This has yet to be used.
        """
        if not file_path.endswith(".xml"):
            raise ValueError("Pass a XML file to return the checksum")
        with open(file_path, encoding="utf-8") as checksum_file:
            checksum_content = checksum_file.read()
        root_elem = etree.fromstring(checksum_content)
        return root_elem.attrib["name"], root_elem.attrib["md5"]

    def generate_installer(self, db_game: Dict[str, Any]) -> Dict[str, Any]:
        details = json.loads(db_game["details"])
        slug = details["slug"]
        platforms = [platform.casefold() for platform, is_supported in details["worksOn"].items() if is_supported]
        if "linux" in platforms:
            return self._generate_installer(slug, "linux", db_game)
        else:
            return self._generate_installer(slug, "wine", db_game)

    def generate_installers(self, db_game: Dict[str, Any]) -> List[dict]:
        details = json.loads(db_game["details"])
        slug = details["slug"]
        platforms = [platform.casefold() for platform, is_supported in details["worksOn"].items() if is_supported]

        installers = []

        if "linux" in platforms:
            installers.append(self._generate_installer(slug, "linux", db_game))

        if "windows" in platforms:
            installers.append(self._generate_installer(slug, "wine", db_game))

        if len(installers) > 1:
            for installer in installers:
                runner_human_name = get_runner_human_name(installer["runner"])
                installer["version"] += " " + (runner_human_name or installer["runner"])

        return installers

    def _generate_installer(self, slug: str, runner: str, db_game: Dict[str, Any]) -> Dict[str, Any]:
        system_config = {}
        if runner == "linux":
            game_config = {"exe": AUTO_ELF_EXE}
            script = [
                {"extract": {"file": "goginstaller", "format": "zip", "dst": "$CACHE"}},
                {"merge": {"src": "$CACHE/data/noarch", "dst": "$GAMEDIR"}},
            ]
        else:
            game_config = {"exe": AUTO_WIN32_EXE}
            script = [
                {"autosetup_gog_game": "goginstaller"},
            ]
        return {
            "name": db_game["name"],
            "version": "GOG",
            "slug": slug,
            "game_slug": self.get_installed_slug(db_game),
            "runner": runner,
            "gogid": db_game["appid"],
            "script": {
                "game": game_config,
                "system": system_config,
                "files": [{"goginstaller": "N/A:Select the installer from GOG"}],
                "installer": script,
            },
        }

    def get_installed_runner_name(self, db_game: dict) -> str:
        platforms = [platform.casefold() for platform in self.get_game_platforms(db_game)]
        return "linux" if "linux" in platforms else "wine"

    def get_games_owned(self) -> dict:
        """Return IDs of games owned by user"""
        url = "{}/user/data/games".format(self.embed_url)
        return self.make_api_request(url)

    def get_dlc_installers(self, db_game: dict) -> List[dict]:
        """Return all available DLC installers for game"""
        appid = db_game["service_id"]
        runner_name = db_game.get("runner")

        filter_os = self.runner_to_os_dict.get(runner_name) if runner_name else None

        dlcs = self.get_game_dlcs(appid)

        installers = []

        for dlc in dlcs:
            dlc_id = "gogdlc-%s" % dlc["slug"]

            # remove mac installers for now
            installfiles = [
                installer for installer in dlc["downloads"].get("installers", []) if installer["os"] != "mac"
            ]

            for file in installfiles:
                file_os = file["os"].casefold()

                if filter_os and file_os and filter_os != file_os:
                    continue

                # supports linux
                if file_os == "linux":
                    runner = "linux"
                    script = [
                        {"extract": {"dst": "$CACHE/GOG", "file": dlc_id, "format": "zip"}},
                        {"merge": {"dst": "$GAMEDIR", "src": "$CACHE/GOG/data/noarch/"}},
                    ]
                else:
                    runner = "wine"
                    script = [{"task": {"name": "wineexec", "executable": dlc_id}}]

                installer = {
                    "name": db_game["name"],
                    # add runner in brackets - wrong installer can be run when this is not unique
                    "version": f"{dlc['title']} ({runner})",
                    "slug": dlc["slug"],
                    "description": "DLC for %s" % db_game["name"],
                    "game_slug": self.get_installed_slug(db_game),
                    "runner": runner,
                    "is_dlc": True,
                    "dlcid": dlc["id"],
                    "gogid": dlc["id"],
                    "script": {
                        "extends": db_game["installer_slug"],
                        "files": [{dlc_id: "N/A:Select the patch from GOG"}],
                        "installer": script,
                    },
                }
                installers.append(installer)

        return installers

    def get_dlc_installers_owned(self, db_game: dict) -> List[dict]:
        """Return DLC installers for owned DLC"""

        owned = self.get_games_owned()
        installers = self.get_dlc_installers(db_game)

        installers = [installer for installer in installers if installer["dlcid"] in owned["owned"]]

        return installers

    def get_dlc_installers_runner(self, db_game: dict, runner: str, only_owned: bool = True) -> List[dict]:
        """Return DLC installers for requested runner
        only_owned=True only return installers for owned DLC (default)"""
        if only_owned:
            installers = self.get_dlc_installers_owned(db_game)
        else:
            installers = self.get_dlc_installers(db_game)

        # only handle linux & wine for now
        if runner != "linux":
            runner = "wine"

        installers = [installer for installer in installers if installer["runner"] == runner]

        return installers

    def get_update_installers(self, db_game: dict) -> List[dict]:
        appid = db_game["service_id"]
        runner = db_game.get("runner")
        patch_versions = self.get_update_versions(appid, runner)
        patch_installers = []
        for version in patch_versions:
            patch = patch_versions[version]
            size = human_size(sum(part["total_size"] for part in patch))
            patch_id = "gogpatch-%s" % slugify(patch[0]["version"])
            installer = {
                "name": db_game["name"],
                "description": patch[0]["name"] + " " + size,
                "slug": db_game["installer_slug"],
                "game_slug": db_game["slug"],
                "version": "GOG " + patch[0]["version"],
                "runner": "wine",
                "script": {
                    "extends": db_game["installer_slug"],
                    "files": [{patch_id: "N/A:Select the patch from GOG"}],
                    "installer": [{"task": {"name": "wineexec", "executable": patch_id}}],
                },
            }
            patch_installers.append(installer)
        return patch_installers

    def get_game_platforms(self, db_game: dict) -> List[str]:
        details = db_game.get("details")
        if details:
            worksOn = json.loads(details).get("worksOn")
            if worksOn is not None:
                return [name for name, works in worksOn.items() if works]
        return []

    def get_game_release_date(self, db_game: dict):
        details = db_game.get("details")
        if details:
            release_date = json.loads(details).get("releaseDate")
            if release_date is not None:
                date = release_date.get("date")
                # GoG stores unknown release dates as a negative date
                if date is not None and isinstance(date, str) and date[0] != "-":
                    # Return as YYYY-MM-DD
                    return date[:10]
        return ""



"""Manipulate MAME ini files"""

# Lutris Modules
from lutris.util.system import path_exists


class MameIni:
    """Looks like an ini file and yet it is not one!"""

    def __init__(self, ini_path):
        if not path_exists(ini_path):
            raise OSError("File %s does not exist" % ini_path)
        self.ini_path = ini_path
        self.lines = []
        self.config = {}

    def parse(self, line):
        """Store configuration value from a line"""
        line = line.strip()
        if not line or line.startswith("#"):
            return None, None
        key, *_value = line.split(maxsplit=1)
        if _value:
            return key, _value[0]
        return key, None

    def read(self):
        """Reads the content of the ini file"""
        with open(self.ini_path, "r", encoding="utf-8") as ini_file:
            for line in ini_file.readlines():
                self.lines.append(line)
                print(line)
                config_key, config_value = self.parse(line)
                if config_key:
                    self.config[config_key] = config_value

    def write(self):
        """Writes the file to disk"""
        with open(self.ini_path, "w", encoding="utf-8") as ini_file:
            for line in self.lines:
                config_key, _value = self.parse(line)
                if config_key and self.config[config_key]:
                    ini_file.write("%-26s%s\n" % (config_key, self.config[config_key]))
                elif config_key:
                    ini_file.write("%s\n" % config_key)
                else:
                    ini_file.write(line)



# pylint: disable=no-member
from gettext import gettext as _
from typing import Sequence

from gi.repository import Gtk

from lutris.database import categories as categories_db
from lutris.database.categories import is_reserved_category
from lutris.game import Game
from lutris.gui.dialogs import QuestionDialog, SavableModelessDialog
from lutris.util.strings import get_natural_sort_key


class EditGameCategoriesDialog(SavableModelessDialog):
    """Game category edit dialog."""

    def __init__(self, game=None, parent=None):
        title = game.name if game else _("Categories")

        super().__init__(title, parent=parent, border_width=10)
        self.set_default_size(350, 250)

        self.category_checkboxes = {}
        self.games = []
        self.categories = sorted(
            [c["name"] for c in categories_db.get_categories() if not is_reserved_category(c["name"])],
            key=lambda c: get_natural_sort_key(c),
        )

        self.checkbox_grid = Gtk.Grid()

        self.vbox.set_homogeneous(False)
        self.vbox.set_spacing(10)
        self.vbox.pack_start(self._create_category_checkboxes(), True, True, 0)
        self.vbox.pack_start(self._create_add_category(), False, False, 0)

        if game:
            self.add_games([game])

        self.vbox.show_all()

    def add_games(self, games: Sequence[Game]) -> None:
        """Adds games to the dialog; this is intended to be used when the dialog is for multiple games,
        and can be used more than once to accumulate games."""

        def mark_category_checkbox(checkbox, included):
            # Checks or unchecks a textbox- but after the first game, this will
            # compare against the current state and go to 'inconsistent' rather than
            # reversing the checkbox.
            if len(self.games) == 0:
                checkbox.set_active(included)
            elif not checkbox.get_inconsistent() and checkbox.get_active() != included:
                checkbox.set_active(False)
                checkbox.set_inconsistent(True)

        def add_game(game):
            # Adds a single game to the dialog, and checks or unchecks
            # boxes as appropriate.
            categories = categories_db.get_categories_in_game(game.id)
            other_checkboxes = set(self.category_checkboxes.values())
            for category in categories:
                category_checkbox = self.category_checkboxes.get(category)
                if category_checkbox:
                    other_checkboxes.discard(category_checkbox)
                    mark_category_checkbox(category_checkbox, included=True)

            for category_checkbox in other_checkboxes:
                mark_category_checkbox(category_checkbox, included=False)

            self.games.append(game)

        existing_game_ids = set(game.id for game in self.games)
        for g in games:
            if g.id not in existing_game_ids:
                add_game(g)

        if len(self.games) > 1:
            subtitle = _("%d games") % len(self.games)
            header_bar = self.get_header_bar()
            if header_bar:
                header_bar.set_subtitle(subtitle)

    def _create_category_checkboxes(self):
        """Constructs a frame containing checkboxes for all known (non-special) categories."""
        frame = Gtk.Frame()
        scrolledwindow = Gtk.ScrolledWindow()

        for category in self.categories:
            label = category
            checkbutton = Gtk.CheckButton(label)
            checkbutton.connect("toggled", self.on_checkbutton_toggled)
            self.checkbox_grid.attach_next_to(checkbutton, None, Gtk.PositionType.BOTTOM, 3, 1)
            self.category_checkboxes[category] = checkbutton

        scrolledwindow.add(self.checkbox_grid)
        frame.add(scrolledwindow)
        return frame

    def _create_add_category(self):
        """Creates a box that carries the controls to add a new category."""

        def on_add_category(*_args):
            category = categories_db.strip_category_name(category_entry.get_text())
            if not categories_db.is_reserved_category(category) and category not in self.category_checkboxes:
                category_entry.set_text("")
                checkbutton = Gtk.CheckButton(category, visible=True, active=True)
                self.category_checkboxes[category] = checkbutton
                self.checkbox_grid.attach_next_to(checkbutton, None, Gtk.PositionType.TOP, 3, 1)
                categories_db.add_category(category)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)

        category_entry = Gtk.Entry()
        category_entry.connect("activate", on_add_category)
        hbox.pack_start(category_entry, True, True, 0)

        button = Gtk.Button.new_with_label(_("Add Category"))
        button.connect("clicked", on_add_category)
        button.set_tooltip_text(_("Adds the category to the list."))
        hbox.pack_end(button, False, False, 0)

        return hbox

    @staticmethod
    def on_checkbutton_toggled(checkbutton):
        # If the user toggles a checkbox, it is no longer inconsistent.
        checkbutton.set_inconsistent(False)

    def on_save(self, _button):
        """Save category changes and destroy widget."""

        changes = []

        for game in self.games:
            for category_checkbox in self.category_checkboxes.values():
                removed_categories = set()
                added_categories = set()

                if not category_checkbox.get_inconsistent():
                    label = category_checkbox.get_label()
                    game_categories = categories_db.get_categories_in_game(game.id)
                    if label in game_categories:
                        if not category_checkbox.get_active():
                            removed_categories.add(label)
                    else:
                        if category_checkbox.get_active():
                            added_categories.add(label)

                if added_categories or removed_categories:
                    changes.append((game, added_categories, removed_categories))

        if changes and len(self.games) > 1:
            if len(changes) == 1:
                question = _("You are updating the categories on 1 game. Are you sure you want to change it?")
            else:
                question = _(
                    "You are updating the categories on %d games. Are you sure you want to change them?"
                ) % len(changes)
            dlg = QuestionDialog(
                {
                    "parent": self,
                    "question": question,
                    "title": _("Changing Categories"),
                }
            )
            if dlg.result != Gtk.ResponseType.YES:
                return

        for game, added_categories, removed_categories in changes:
            game.update_game_categories(added_categories, removed_categories)

        self.destroy()



import os
from gettext import gettext as _
from typing import Optional
from urllib.parse import urlparse

from gi.repository import GObject, Gtk, Pango

from lutris.gui.dialogs import display_error
from lutris.util.downloader import Downloader
from lutris.util.jobs import schedule_repeating_at_idle
from lutris.util.log import logger
from lutris.util.strings import gtk_safe, human_size


class DownloadProgressBox(Gtk.Box):
    """Progress bar used to monitor a file download."""

    __gsignals__ = {
        "complete": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
        "cancel": (GObject.SignalFlags.RUN_LAST, None, ()),
        "error": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(
        self,
        url: str,
        dest: str,
        temp: str = None,
        referer: Optional[str] = None,
        title: Optional[str] = None,
        cancelable: bool = True,
        downloader: Optional[Downloader] = None,
    ) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self._downloader = downloader
        self.is_complete = False
        self.url = url
        self.dest = dest
        self.temp = temp or (dest + ".tmp")
        self.referer = referer

        if not title:
            parsed_url = urlparse(url)
            title = "%s%s" % (parsed_url.netloc, parsed_url.path)

        self.main_label = Gtk.Label(label=title)
        self.main_label.set_alignment(0, 0)
        self.main_label.set_property("wrap", True)
        self.main_label.set_margin_bottom(10)
        # self.main_label.set_max_width_chars(70)
        self.main_label.set_selectable(True)
        self.main_label.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        self.pack_start(self.main_label, True, True, 0)

        progress_box = Gtk.Box()

        self.progressbar = Gtk.ProgressBar()
        self.progressbar.set_margin_top(5)
        self.progressbar.set_margin_bottom(5)
        self.progressbar.set_margin_right(10)
        progress_box.pack_start(self.progressbar, True, True, 0)

        self.cancel_button = Gtk.Button.new_with_mnemonic(_("_Cancel"))
        self.cancel_cb_id = self.cancel_button.connect("clicked", self.on_cancel_clicked)
        if not cancelable:
            self.cancel_button.set_sensitive(False)
        progress_box.pack_end(self.cancel_button, False, False, 0)

        self.pack_start(progress_box, False, False, 0)

        self.progress_label = Gtk.Label()
        self.progress_label.set_alignment(0, 0)
        self.pack_start(self.progress_label, True, True, 0)

        self.show_all()
        self.cancel_button.hide()

        if os.path.exists(self.temp):
            os.remove(self.temp)

    @property
    def downloader(self) -> Downloader:
        if not self._downloader:
            self._downloader = Downloader(self.url, self.temp, referer=self.referer, overwrite=True)
        return self._downloader

    def cancel_download(self):
        if self._downloader:
            self._downloader.cancel()

    def start(self) -> None:
        """Start downloading a file."""
        try:
            downloader = self.downloader
        except RuntimeError as ex:
            display_error(ex, parent=self.get_toplevel())
            self.emit("cancel")
            return None

        schedule_repeating_at_idle(self._progress, interval_seconds=0.5)
        self.cancel_button.show()
        self.cancel_button.set_sensitive(True)
        if not downloader.state == downloader.DOWNLOADING:
            downloader.start()

    def set_retry_button(self):
        """Transform the cancel button into a retry button"""
        self.cancel_button.set_label(_("Retry"))
        self.cancel_button.disconnect(self.cancel_cb_id)
        self.cancel_cb_id = self.cancel_button.connect("clicked", self.on_retry_clicked)
        self.cancel_button.set_sensitive(True)

    def on_retry_clicked(self, button):
        logger.debug("Retrying download")
        button.set_label(_("Cancel"))
        button.disconnect(self.cancel_cb_id)
        self.cancel_cb_id = button.connect("clicked", self.on_cancel_clicked)
        self.downloader.reset()
        self.start()

    def on_cancel_clicked(self, _widget=None):
        """Cancel the current download."""
        logger.debug("Download cancel requested")
        self.cancel_download()
        self.cancel_button.set_sensitive(False)
        self.emit("cancel")

    def _progress(self) -> bool:
        """Show download progress."""
        downloader = self.downloader
        progress = min(downloader.check_progress(), 1)
        if downloader.state in [downloader.CANCELLED, downloader.ERROR]:
            self.progressbar.set_fraction(0)
            if downloader.state == downloader.CANCELLED:
                self._set_text(_("Download interrupted"))
                self.emit("cancel")
            else:
                self._set_text(str(downloader.error)[:80])
                self.emit("error", downloader.error)
            return False
        self.progressbar.set_fraction(progress)
        megabytes = 1024 * 1024
        progress_text = _("{downloaded} / {size} ({speed:0.2f}MB/s), {time} remaining").format(
            downloaded=human_size(downloader.downloaded_size),
            size=human_size(downloader.full_size),
            speed=float(downloader.average_speed) / megabytes,
            time=downloader.time_left,
        )
        self._set_text(progress_text)
        if downloader.state == downloader.COMPLETED:
            os.rename(self.temp, self.dest)
            self.cancel_button.set_sensitive(False)
            self.is_complete = True
            self.emit("complete", {})
            return False
        return True

    def _set_text(self, text):
        markup = "<span size='10000'>{}</span>".format(gtk_safe(text))
        self.progress_label.set_markup(markup)



from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system


class xemu(Runner):
    human_name = _("xemu")
    platforms = [_("Xbox")]
    description = _("Xbox emulator")
    runnable_alone = True
    runner_executable = "xemu/xemu"
    flatpak_id = "app.xemu.xemu"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ISO file"),
            "help": _("DVD image in iso format"),
        }
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "label": _("Fullscreen"),
            "type": "bool",
            "default": True,
        },
    ]

    # xemu currently uses an AppImage, no need for the runtime.
    system_options_override = [{"option": "disable_runtime", "default": True}]

    def play(self):
        """Run the game."""
        arguments = self.get_command()

        fullscreen = self.runner_config.get("fullscreen")
        if fullscreen:
            arguments.append("-full-screen")

        iso = self.game_config.get("main_file") or ""
        if not system.path_exists(iso):
            raise MissingGameExecutableError(filename=iso)
        arguments += ["-dvd_path", iso]
        return {"command": arguments}



from gettext import gettext as _
from typing import Any, Dict, Optional

from gi.repository import Gio, Gtk  # type: ignore

from lutris import settings
from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.config.widget_generator import WidgetGenerator
from lutris.gui.widgets.status_icon import supports_status_icon
from lutris.settings import read_setting


def _is_system_dark_by_default():
    app = Gio.Application.get_default()
    return app.style_manager.is_dark_by_default


class InterfacePreferencesBox(BaseConfigBox):
    settings_options = [
        {
            "option": "hide_client_on_game_start",
            "label": _("Minimize client when a game is launched"),
            "type": "bool",
            "help": _("Minimize the Lutris window while playing a game; it will return when the game exits."),
        },
        {
            "option": "hide_text_under_icons",
            "label": _("Hide text under icons"),
            "type": "bool",
            "help": _("Removes the names from the Lutris window when in grid view, but not list view."),
        },
        {
            "option": "hide_badges_on_icons",
            "label": _("Hide badges on icons (Ctrl+p to toggle)"),
            "type": "bool",
            "accelerator": "<Primary>p",
            "help": _("Removes the platform and missing-game badges from icons in the Lutris window."),
        },
        {
            "option": "show_tray_icon",
            "label": _("Show Tray Icon"),
            "type": "bool",
            "available": supports_status_icon,
            "help": _(
                "Adds a Lutris icon to the tray, and prevents Lutris from exiting when the Lutris window is closed. "
                "You can still exit using the menu of the tray icon."
            ),
        },
        {
            "option": "discord_rpc",
            "label": _("Enable Discord Rich Presence for Available Games"),
            "type": "bool",
        },
        {
            "option": "preferred_theme",
            "type": "choice",
            "label": _("Theme"),
            "choices": [
                (_("System Default"), "default"),
                (_("Light"), "light"),
                (_("Dark"), "dark"),
            ],
            "default": "default",
            "help": _("Overrides Lutris's appearance to be light or dark."),
        },
    ]

    def __init__(self, accelerators):
        super().__init__()
        self.accelerators = accelerators

        self.add(self.get_section_label(_("Interface options")))
        frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)
        listbox = Gtk.ListBox(visible=True)
        frame.add(listbox)
        self.pack_start(frame, False, False, 0)

        gen = PreferencesWidgetGenerator(listbox)
        gen.changed.register(self.on_setting_changed)
        self.widget_generator = gen

        for option in self.settings_options:
            gen.generate_container(option)

            if gen.option_container:
                list_box_row = Gtk.ListBoxRow(visible=True)
                list_box_row.set_selectable(False)
                list_box_row.set_activatable(False)
                list_box_row.add(gen.option_container)
                listbox.add(list_box_row)

        gen.update_widgets()

    def on_setting_changed(self, option_key, new_value):
        settings.write_setting(option_key, new_value)


class PreferencesWidgetGenerator(WidgetGenerator):
    """This generator adjusts the spacing of the wrappers and packs widgets on the
    right to get the interface preferences layout instead of the configuration one."""

    def get_setting(self, option_key: str, default: Any) -> Any:
        return read_setting(option_key, default=default)

    def create_wrapper_box(self, option: Dict[str, Any], value: Any, default: Any) -> Optional[Gtk.Box]:
        box = super().create_wrapper_box(option, value, default)
        if box:
            box.set_margin_top(12)
            box.set_margin_bottom(12)
            box.set_margin_right(12)
            box.set_margin_left(12)
        return box

    def build_option_widget(
        self, option: Dict[str, Any], widget: Optional[Gtk.Widget], no_label: bool = False, expand: bool = False
    ) -> Optional[Gtk.Widget]:
        if no_label:
            return super().build_option_widget(option, widget, no_label=no_label, expand=expand)

        label = Gtk.Label(option["label"], visible=True, wrap=True)
        label.set_alignment(0, 0.5)
        if self.wrapper and widget:
            self.wrapper.pack_start(label, True, True, 0)
            self.wrapper.pack_end(widget, expand, expand, 0)
        return widget



"""Threading module, used to launch games while monitoring them."""

import contextlib
import fcntl
import io
import os
import shlex
import subprocess
import sys
import uuid
from copy import copy
from typing import List

from gi.repository import GLib

from lutris import settings
from lutris.util import system
from lutris.util.log import logger
from lutris.util.shell import get_terminal_script


def get_wrapper_script_location():
    """Return absolute path of lutris-wrapper script"""
    wrapper_relpath = "share/lutris/bin/lutris-wrapper"
    candidates = [
        os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), "..")),
        os.path.dirname(os.path.dirname(settings.__file__)),
        "/usr",
        "/usr/local",
    ]
    for candidate in candidates:
        wrapper_abspath = os.path.join(candidate, wrapper_relpath)
        if os.path.isfile(wrapper_abspath):
            return wrapper_abspath
    raise FileNotFoundError("Couldn't find lutris-wrapper script in any of the expected locations")


WRAPPER_SCRIPT = get_wrapper_script_location()
RUNNING_COMMANDS = set()


class MonitoredCommand:
    """Exexcutes a commmand while keeping track of its state"""

    fallback_cwd = "/tmp"

    def __init__(
        self,
        command,
        runner=None,
        env=None,
        term=None,
        cwd=None,
        include_processes=None,
        exclude_processes=None,
        log_buffer=None,
        title=None,
    ):  # pylint: disable=too-many-arguments
        self.ready_state = True
        self.env = self.get_environment(env)

        self.accepted_return_code = "0"

        self.command = command
        self.runner = runner
        self.stop_func = lambda: True
        self.game_process = None
        self.prevent_on_stop = False
        self.return_code = None
        self.terminal = term
        self.is_running = True
        self.error = None
        self.log_handlers = [
            self.log_handler_stdout,
            self.log_handler_console_output,
        ]
        self.set_log_buffer(log_buffer)
        self.stdout_monitor = None
        self.include_processes = include_processes or []
        self.exclude_processes = exclude_processes or []

        self.cwd = self.get_cwd(cwd)

        self._stdout = io.StringIO()

        self._title = title if title else command[0]

    @property
    def stdout(self):
        return self._stdout.getvalue()

    def get_wrapper_command(self) -> List[str]:
        """Return launch arguments for the wrapper script"""
        wrapper_command = (
            [
                WRAPPER_SCRIPT,
                self._title,
                str(len(self.include_processes)),
                str(len(self.exclude_processes)),
            ]
            + self.include_processes
            + self.exclude_processes
        )
        if not self.terminal:
            return wrapper_command + self.command

        terminal_path = system.find_required_executable(self.terminal)
        script_path = get_terminal_script(self.command, self.cwd, self.env)
        return wrapper_command + [terminal_path, "-e", script_path]

    def set_log_buffer(self, log_buffer):
        """Attach a TextBuffer to this command enables the buffer handler"""
        if not log_buffer:
            return
        self.log_buffer = log_buffer
        if self.log_handler_buffer not in self.log_handlers:
            self.log_handlers.append(self.log_handler_buffer)

    def get_cwd(self, cwd):
        """Return the current working dir of the game"""
        if not cwd:
            cwd = self.runner.working_dir if self.runner else None
        return os.path.expanduser(cwd or "~")

    @staticmethod
    def get_environment(user_env):
        """Process the user provided environment variables for use as self.env"""
        env = copy(user_env) if user_env else {}

        # not clear why this needs to be added, the path is already added in
        # the wrappper script.
        env["PYTHONPATH"] = ":".join(sys.path)
        # Drop bad values of environment keys, those will confuse the Python
        # interpreter.
        env["LUTRIS_GAME_UUID"] = str(uuid.uuid4())

        cleaned = {}
        for key, value in env.items():
            if "=" in key:
                logger.warning("Environment variable name '%s' contains '=' so it can't be used; skipping.", key)
            elif value is None:
                logger.warning("Environment variable '%s' has None for its value; skipping.", key)
            elif not isinstance(value, str):
                logger.warning("Environment variable '%s' value '%s' is not a string; converting.", key, value)
                cleaned[key] = str(value)
            else:
                cleaned[key] = value
        return cleaned

    def get_child_environment(self):
        """Returns the calculated environment for the child process."""
        env = system.get_environment()
        env.update(self.env)
        return env

    def start(self):
        """Run the thread."""
        if os.environ.get("LUTRIS_DEBUG_ENV") == "1":
            for key, value in self.env.items():
                logger.debug('%s="%s"', key, value)
        wrapper_command = self.get_wrapper_command()
        env = self.get_child_environment()

        self.game_process = self.execute_process(wrapper_command, env)
        RUNNING_COMMANDS.add(self)

        if not self.game_process:
            logger.error("No game process available")
            return

        GLib.child_watch_add(self.game_process.pid, self.on_stop)

        # make stdout nonblocking.
        fileno = self.game_process.stdout.fileno()
        fcntl.fcntl(fileno, fcntl.F_SETFL, fcntl.fcntl(fileno, fcntl.F_GETFL) | os.O_NONBLOCK)

        self.stdout_monitor = GLib.io_add_watch(
            self.game_process.stdout,
            GLib.IO_IN | GLib.IO_HUP,
            self.on_stdout_output,
        )

    def log_filter(self, line: str) -> bool:
        """Filter out some message we don't want to show to the user."""
        if "GStreamer-WARNING **" in line:
            return False
        if "Bad file descriptor" in line:
            return False
        if "'libgamemodeauto.so.0' from LD_PRELOAD" in line:
            return False
        if "Unable to read VR Path Registry" in line:
            return False
        return True

    def log_handler_stdout(self, line):
        """Add the line to this command's stdout attribute"""
        if not self.log_filter(line):
            return
        self._stdout.write(line)

    def log_handler_buffer(self, line):
        """Add the line to the associated LogBuffer object"""
        self.log_buffer.insert(self.log_buffer.get_end_iter(), line, -1)

    def log_handler_console_output(self, line):
        """Print the line to stdout"""
        if not self.log_filter(line):
            return
        with contextlib.suppress(BlockingIOError):
            sys.stdout.write(line)
            sys.stdout.flush()

    def get_return_code(self):
        """Get the return code from the file written by the wrapper"""
        return_code_path = "/tmp/lutris-%s" % self.env["LUTRIS_GAME_UUID"]
        if os.path.exists(return_code_path):
            with open(return_code_path, encoding="utf-8") as return_code_file:
                return_code = return_code_file.read()
            os.unlink(return_code_path)
        else:
            return_code = ""
            logger.warning("No file %s", return_code_path)
        return return_code

    def on_stop(self, pid, _user_data):
        """Callback registered on game process termination"""
        if self.prevent_on_stop:  # stop() already in progress
            return False
        self.game_process.wait()
        self.return_code = self.get_return_code()
        self.is_running = False
        logger.debug("Process %s has terminated with code %s", pid, self.return_code)
        resume_stop = self.stop()
        if not resume_stop:
            logger.info("Full shutdown prevented")
            return False
        return False

    def on_stdout_output(self, stdout, condition):
        """Called by the stdout monitor to dispatch output to log handlers"""
        if condition == GLib.IO_HUP:
            self.stdout_monitor = None
            return False
        if not self.is_running:
            return False
        try:
            line = stdout.read(262144).decode("utf-8", errors="ignore")
        except ValueError:
            # file_desc might be closed
            return True
        if "winemenubuilder.exe" in line:
            return True
        for log_handler in self.log_handlers:
            log_handler(line)
        return True

    def execute_process(self, command, env=None):
        """Execute and return a subprocess"""

        # If a None gets into execute_process, we get annoying errors
        # that are hard to race. We'll try to repair the bad command or environment
        # instead, while emitting warnings.abs

        for i, item in enumerate(command):
            if not isinstance(item, str):
                logger.warning("Wrapper command contains a non-string: %s", command)
                command[i] = str(item) if item else ""

        if "" in env:
            del env[""]

        for key, value in env.items():
            if not isinstance(key, str) or key.isspace():
                logger.warning("Environment contains a non-string as a key %s=%s: %s", key, value, env)
                env = copy(env)  # can't del while iterating
                del env[key]
                continue

            if not isinstance(value, str):
                logger.warning("Environment contains a non-string as the value for the key: %s=%s: %s", key, value, env)
                env[key] = str(value) if value else ""

        if self.cwd and not system.path_exists(self.cwd):
            try:
                os.makedirs(self.cwd)
            except OSError:
                logger.error("Failed to create working directory, falling back to %s", self.fallback_cwd)
                self.cwd = "/tmp"
        try:
            return subprocess.Popen(  # pylint: disable=consider-using-with
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd=self.cwd,
                env=env,
            )
        except OSError as ex:
            logger.exception("Failed to execute %s: %s", " ".join(command), ex)
            self.error = ex.strerror

    def stop(self):
        """Stops the current game process and cleans up the instance"""
        # Prevent stop() being called again by the process exiting
        self.prevent_on_stop = True

        try:
            self.game_process.terminate()
        except ProcessLookupError:
            # process already dead.
            pass

        resume_stop = self.stop_func()
        if not resume_stop:
            logger.warning("Stop execution halted by demand of stop_func")
            return False

        if self.stdout_monitor:
            GLib.source_remove(self.stdout_monitor)
            self.stdout_monitor = None

        self.is_running = False
        self.ready_state = False
        RUNNING_COMMANDS.discard(self)
        return True


def exec_command(command):
    """Execute arbitrary command in a MonitoredCommand

    Used by the --exec command line flag.
    """
    command = MonitoredCommand(shlex.split(command), env={})  # runtime.get_env())
    command.start()
    return command



from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system

VER_STR = "2123.2"


class azahar(Runner):
    human_name = _("Azahar")
    platforms = [_("Nintendo 3DS")]
    description = _("Nintendo 3DS Emulator")
    runnable_alone = True
    runner_executable = "azahar/azahar.AppImage"
    flatpak_id = "org.azahar_emu.Azahar"
    download_url = f"https://github.com/azahar-emu/azahar/releases/download/{VER_STR}/azahar.AppImage"

    # Azahar uses an AppImage, runtime causes QT  platform plugins issues.
    system_options_override = [{"option": "disable_runtime", "default": True}]
    game_options = [
        {
            "option": "main_file",
            "label": _("The game data, commonly called a ROM image."),
            "type": "file",
            "default_path": "game_path",
        },
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "label": _("Fullscreen"),
            "type": "bool",
            "default": True,
        },
    ]

    def play(self):
        """Run the game."""
        arguments = self.get_command()
        rom = self.game_config.get("main_file") or ""

        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)

        fullscreen = self.runner_config.get("fullscreen") or ""

        if fullscreen:
            arguments.append("-f")
        # Don't pass '-w' here if `fullscreen' is not set to avoid interfering
        # with (possible) present of future emulator caching of last used mode.
        # Like that we avoid complications of having both `fullscreen' and
        # `windowed' options to handle.

        arguments.append(rom)
        return {"command": arguments}



# pylint: disable=no-member
from copy import copy
from gettext import gettext as _
from typing import Callable, Dict, List, Optional, Tuple

from gi.repository import GObject, Gtk

from lutris import runners, services
from lutris.database import categories as categories_db
from lutris.database import games as games_db
from lutris.database import saved_searches
from lutris.database.saved_searches import SavedSearch
from lutris.exceptions import InvalidSearchTermError
from lutris.gui.dialogs import QuestionDialog, SavableModelessDialog
from lutris.gui.widgets.utils import has_stock_icon
from lutris.search import FLAG_TEXTS, GameSearch
from lutris.search_predicate import AndPredicate, SearchPredicate, format_flag

DEFAULT_NEW_SEARCH_NAME = _("New Saved Search")


class SearchFiltersBox(Gtk.Box):
    """A widget to edit dynamic categories"""

    __gsignals__ = {
        "saved": (GObject.SIGNAL_RUN_FIRST, None, (str,)),
    }

    def __init__(self, saved_search: SavedSearch, search_entry: Gtk.SearchEntry = None, can_save: bool = True) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.saved_search = copy(saved_search)
        self.original_search = copy(saved_search)

        if not self.saved_search.name:
            self.saved_search.name = DEFAULT_NEW_SEARCH_NAME

        self.search = self.saved_search.search

        self.set_homogeneous(False)
        self.set_margin_top(20)
        self.set_margin_bottom(20)
        self.set_margin_start(20)
        self.set_margin_end(20)
        self.set_spacing(10)

        self.name_entry = self._add_entry_box(
            _("Name"),
            self.saved_search.name,
            ["tag-symbolic", "poi-marker", "favorite-symbolic"] if can_save else None,
            self.on_save,
        )

        self.search_entry = search_entry or self._add_entry_box(_("Search"), self.search)
        self.search_entry.connect("changed", self.on_search_entry_changed)

        self.predicate_widget_functions: Dict[str, Callable[[SearchPredicate], None]] = {}
        self.updating_predicate_widgets = False

        predicates_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        self.flags_grid = Gtk.Grid(row_spacing=6, column_spacing=6)
        self._add_flag_widgets()

        categories_scrolled_window = Gtk.ScrolledWindow(visible=True)
        categories_scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        categories_frame = Gtk.Frame(visible=True)
        categories_frame.get_style_context().add_class("info-frame")
        categories_frame.add(categories_scrolled_window)
        self.categories_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        categories_scrolled_window.add(self.categories_box)
        categories_frame_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        categories_frame_box.pack_start(Gtk.Label(label=_("Categories"), halign=Gtk.Align.START), False, False, 0)
        categories_frame_box.pack_start(categories_frame, True, True, 0)

        self._add_category_widgets()

        self.update_predicate_widgets()

        predicates_box.pack_start(self.flags_grid, False, False, 0)
        predicates_box.pack_start(categories_frame_box, True, True, 0)
        self.pack_start(predicates_box, True, True, 0)

        self.show_all()

    def _add_entry_box(
        self, label: str, text: str, button_icon_names: Optional[List[str]] = None, clicked: Optional[Callable] = None
    ) -> Gtk.Entry:
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        entry_label = Gtk.Label(label=label)
        entry_label.set_alignment(0, 0.5)
        entry_label.set_size_request(120, -1)
        entry = Gtk.Entry()
        entry.set_text(text)
        hbox.pack_start(entry_label, False, False, 0)
        hbox.pack_start(entry, True, True, 0)

        if button_icon_names:
            button_icon_names = [name for name in button_icon_names if has_stock_icon(name)]
            if button_icon_names:
                button = Gtk.Button.new_from_icon_name(button_icon_names[0], Gtk.IconSize.BUTTON)
                button.get_style_context().add_class("circular")
                if clicked:
                    button.connect("clicked", clicked)
                hbox.pack_end(button, False, False, 0)

        self.pack_start(hbox, False, False, 0)
        return entry

    def on_search_entry_changed(self, _widget):
        self.update_predicate_widgets()

    def update_predicate_widgets(self):
        if not self.updating_predicate_widgets:
            try:
                self.updating_predicate_widgets = True
                self.search = self.search_entry.get_text()
                predicate = GameSearch(self.search).get_predicate()

                for _control, func in self.predicate_widget_functions.items():
                    func(predicate)
            except InvalidSearchTermError:
                pass
            finally:
                self.updating_predicate_widgets = False

    def _add_flag_widgets(self):
        self._add_service_widget(0)
        self._add_runner_widget(1)
        self._add_platform_widget(2)
        self._add_flag_widget(3, _("Installed"), "installed")
        self._add_flag_widget(4, _("Favorite"), "favorite")
        self._add_flag_widget(5, _("Hidden"), "hidden")
        self._add_flag_widget(6, _("Categorized"), "categorized")

    def _change_search_flag(self, tag: str, flag: Optional[bool]):
        game_search = GameSearch(self.search)
        flag_predicate = game_search.get_flag_predicate(tag, flag)
        if flag_predicate:
            predicate = game_search.get_predicate().without_flag(tag)
            predicate = AndPredicate([predicate, flag_predicate]).simplify()
            self.search = str(predicate)
            self.search_entry.set_text(self.search)

    def _remove_search_flag(self, tag: str):
        predicate = GameSearch(self.search).get_predicate().without_flag(tag)
        self.search = str(predicate)
        self.search_entry.set_text(self.search)

    def _add_flag_widget(self, row, caption, tag):
        def on_combobox_change(_widget):
            if not self.updating_predicate_widgets:
                active_id = combobox.get_active_id()
                if active_id == "":
                    self._remove_search_flag(tag)
                else:
                    self._change_search_flag(tag, FLAG_TEXTS[active_id])

        def populate_widget(predicate):
            if predicate.has_flag(tag):
                combobox.set_active_id(format_flag(predicate.get_flag(tag)))
            else:
                combobox.set_active_id("")

        label = Gtk.Label(caption, halign=Gtk.Align.START, valign=Gtk.Align.CENTER)
        label.set_alignment(0, 0.5)
        label.set_size_request(120, -1)
        self.flags_grid.attach(label, 0, row, 1, 1)

        options = [
            (_("-"), ""),
            (_("Yes"), "yes"),
            (_("No"), "no"),
        ]

        combobox = self._create_combobox(options)
        self.predicate_widget_functions[combobox] = populate_widget
        self.flags_grid.attach(combobox, 1, row, 1, 1)
        combobox.connect("changed", on_combobox_change)

    def _add_service_widget(self, row):
        options = [(s[1]().name, s[0]) for s in services.get_enabled_services().items()]
        options.append(("(none)", "none"))

        self._add_match_widget(
            row, _("Source"), "source", options, predicate_factory=lambda s, v: s.get_service_predicate(v)
        )

    def _add_runner_widget(self, row):
        options = [(r.human_name, r.name) for r in runners.get_installed()]
        options.append(("(none)", "none"))

        self._add_match_widget(
            row, _("Runner"), "runner", options, predicate_factory=lambda s, v: s.get_runner_predicate(v)
        )

    def _add_platform_widget(self, row):
        options = [(p, p) for p in games_db.get_used_platforms()]
        options.append(("(none)", "none"))

        self._add_match_widget(
            row, _("Platform"), "platform", options, predicate_factory=lambda s, v: s.get_platform_predicate(v)
        )

    def _add_match_widget(self, row: int, caption: str, tag: str, options: List[Tuple[str, str]], predicate_factory):
        def on_combobox_change(_widget):
            if not self.updating_predicate_widgets:
                game_search = GameSearch(self.search)
                predicate = game_search.get_predicate().without_match(tag)
                active_id = combobox.get_active_id()
                if active_id:
                    p = predicate_factory(game_search, active_id)
                    predicate = AndPredicate([predicate, p]).simplify()
                self.search = str(predicate)
                self.search_entry.set_text(self.search)

        def populate_widget(predicate):
            matches = predicate.get_matches(tag)
            if matches:
                combobox.set_active_id(matches[0])
            else:
                combobox.set_active_id("")

        label = Gtk.Label(label=caption, halign=Gtk.Align.START, valign=Gtk.Align.CENTER)
        self.flags_grid.attach(label, 0, row, 1, 1)

        options = [(_("-"), "")] + options
        combobox = self._create_combobox(options)
        self.predicate_widget_functions[combobox] = populate_widget
        self.flags_grid.attach(combobox, 1, row, 1, 1)
        combobox.connect("changed", on_combobox_change)

    def _add_category_widgets(self):
        for category in categories_db.get_categories():
            category_name = category["name"]
            if not categories_db.is_reserved_category(category_name):
                self._add_category_widget(category_name, category_name)

    def _add_category_widget(self, caption, category_name):
        checkbox = Gtk.CheckButton(caption)

        def on_checkbox_toggled(_widget):
            if not self.updating_predicate_widgets:
                game_search = GameSearch(self.search)
                predicate = game_search.get_predicate().without_match("category", category_name)
                if checkbox.get_active():
                    category_predicate = game_search.get_category_predicate(category_name)
                    predicate = AndPredicate([predicate, category_predicate]).simplify()
                self.search = str(predicate)
                self.search_entry.set_text(self.search)

        def populate_widget(predicate):
            matched = category_name in predicate.get_matches("category")
            checkbox.set_active(matched)

        self.predicate_widget_functions[checkbox] = populate_widget
        self.categories_box.pack_start(checkbox, False, False, 0)
        checkbox.connect("toggled", on_checkbox_toggled)

    def _create_combobox(self, options):
        liststore = Gtk.ListStore(str, str)

        for option in options:
            liststore.append(option)

        combobox = Gtk.ComboBox.new_with_model(liststore)
        combobox.set_entry_text_column(0)
        combobox.set_id_column(1)
        combobox.set_halign(Gtk.Align.START)
        combobox.set_valign(Gtk.Align.CENTER)
        combobox.set_size_request(240, -1)
        renderer_text = Gtk.CellRendererText()
        combobox.pack_start(renderer_text, True)
        combobox.add_attribute(renderer_text, "text", 0)
        return combobox

    @property
    def search_name(self):
        name = self.name_entry.get_text() or self.original_search.name
        return saved_searches.strip_saved_search_name(name)

    @search_name.setter
    def search_name(self, value):
        value = saved_searches.strip_saved_search_name(value)
        self.name_entry.set_text(value)
        self.saved_search.name = value

    def on_save(self, _button: Gtk.Button) -> None:
        """Save game info and destroy widget."""
        search_name = self.search_name
        self.saved_search.name = search_name
        self.saved_search.search = str(GameSearch(self.search_entry.get_text()))

        if self.original_search.name != self.saved_search.name:
            if saved_searches.get_saved_search_by_name(self.saved_search.name):
                raise RuntimeError(_("'%s' is already a saved search.") % self.saved_search.name)

        if not self.saved_search.saved_search_id:
            # Creating new search!
            self.saved_search.add()
        elif self.original_search != self.saved_search:
            # Changing an existing search.
            self.saved_search.update()

        self.search_name = DEFAULT_NEW_SEARCH_NAME
        self.emit("saved", search_name)


class EditSavedSearchDialog(SavableModelessDialog):
    """A dialog to edit saved searches."""

    def __init__(self, parent, saved_search: SavedSearch) -> None:
        self.filter_box = SearchFiltersBox(saved_search, can_save=False)
        self.saved_search = copy(saved_search)
        self.original_search = copy(saved_search)

        if not self.saved_search.name:
            self.saved_search.name = DEFAULT_NEW_SEARCH_NAME
        title = _("Configure %s") % self.saved_search.name
        super().__init__(title, parent=parent, border_width=10)
        self.set_default_size(600, -1)

        content_area = self.get_content_area()
        content_area.set_homogeneous(False)
        content_area.set_spacing(10)
        content_area.pack_start(self.filter_box, True, True, 0)

        delete_button = self.add_styled_button(Gtk.STOCK_DELETE, Gtk.ResponseType.NONE, css_class="destructive-action")
        delete_button.connect("clicked", self.on_delete_clicked)
        delete_button.show() if self.saved_search.saved_search_id else delete_button.hide()

    def on_save(self, button: Gtk.Button) -> None:
        """Save game info and destroy widget."""
        self.filter_box.on_save(button)
        self.destroy()

    def on_delete_clicked(self, _button):
        dlg = QuestionDialog(
            {
                "title": _("Do you want to delete the saved search '%s'?") % self.original_search.name,
                "question": _(
                    "This will permanently destroy the saved search, but the games themselves will not be deleted."
                ),
                "parent": self,
            }
        )
        if dlg.result == Gtk.ResponseType.YES:
            self.saved_search.remove()
            self.destroy()



"""
Module for deserializing/serializing to and from VDF

https://github.com/ValvePython/vdf

MIT License
"""
# pylint: disable=raise-missing-from

__version__ = "3.2"
__author__ = "Rossen Georgiev"

import re
import struct
from binascii import crc32
from io import StringIO as unicodeIO

string_type = str
int_type = int
BOMS = "\ufffe\ufeff"


def strip_bom(line):
    return line.lstrip(BOMS)


# string escaping
_unescape_char_map = {
    r"\n": "\n",
    r"\t": "\t",
    r"\v": "\v",
    r"\b": "\b",
    r"\r": "\r",
    r"\f": "\f",
    r"\a": "\a",
    r"\\": "\\",
    r"\?": "?",
    r"\"": '"',
    r"\'": "'",
}
_escape_char_map = {v: k for k, v in _unescape_char_map.items()}


def _re_escape_match(m):
    return _escape_char_map[m.group()]


def _re_unescape_match(m):
    return _unescape_char_map[m.group()]


def _escape(text):
    return re.sub(r"[\n\t\v\b\r\f\a\\\?\"']", _re_escape_match, text)


def _unescape(text):
    return re.sub(r"(\\n|\\t|\\v|\\b|\\r|\\f|\\a|\\\\|\\\?|\\\"|\\')", _re_unescape_match, text)


# parsing and dumping for KV1


def parse(fp, mapper=dict, merge_duplicate_keys=True, escaped=True):
    """
    Deserialize ``s`` (a ``str`` or ``unicode`` instance containing a VDF)
    to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.
    """
    if not issubclass(mapper, dict):
        raise TypeError("Expected mapper to be subclass of dict, got %s" % type(mapper))
    if not hasattr(fp, "readline"):
        raise TypeError("Expected fp to be a file-like object supporting line iteration")

    lineno = 0
    stack = [mapper()]
    expect_bracket = False

    re_keyvalue = re.compile(
        r'^("(?P<qkey>(?:\\.|[^\\"])+)"|(?P<key>#?[a-z0-9\-\_\\\?]+))'
        r"([ \t]*("
        r'"(?P<qval>(?:\\.|[^\\"])*)(?P<vq_end>")?'
        r"|(?P<val>[a-z0-9\-\_\\\?\*\.]+)"
        r"))?",
        flags=re.I,
    )

    for lineno, line in enumerate(fp, 1):
        if lineno == 1:
            line = strip_bom(line)

        line = line.lstrip()

        # skip empty and comment lines
        if line == "" or line[0] == "/":
            continue

        # one level deeper
        if line[0] == "{":
            expect_bracket = False
            continue

        if expect_bracket:
            raise SyntaxError(
                "vdf.parse: expected openning bracket",
                (getattr(fp, "name", "<%s>" % fp.__class__.__name__), lineno, 1, line),
            )

        # one level back
        if line[0] == "}":
            if len(stack) > 1:
                stack.pop()
                continue

            raise SyntaxError(
                "vdf.parse: one too many closing parenthasis",
                (getattr(fp, "name", "<%s>" % fp.__class__.__name__), lineno, 0, line),
            )

        # parse keyvalue pairs
        while True:
            match = re_keyvalue.match(line)

            if not match:
                try:
                    line += next(fp)
                    continue
                except StopIteration:
                    raise SyntaxError(
                        "vdf.parse: unexpected EOF (open key quote?)",
                        (getattr(fp, "name", "<%s>" % fp.__class__.__name__), lineno, 0, line),
                    )

            key = match.group("key") if match.group("qkey") is None else match.group("qkey")
            val = match.group("val") if match.group("qval") is None else match.group("qval")

            if escaped:
                key = _unescape(key)

            # we have a key with value in parenthesis, so we make a new dict obj (level deeper)
            if val is None:
                if merge_duplicate_keys and key in stack[-1]:
                    _m = stack[-1][key]
                else:
                    _m = mapper()
                    stack[-1][key] = _m

                stack.append(_m)
                expect_bracket = True

            # we've matched a simple keyvalue pair, map it to the last dict obj in the stack
            else:
                # if the value is line consume one more line and try to match again,
                # until we get the KeyValue pair
                if match.group("vq_end") is None and match.group("qval") is not None:
                    try:
                        line += next(fp)
                        continue
                    except StopIteration:
                        raise SyntaxError(
                            "vdf.parse: unexpected EOF (open quote for value?)",
                            (getattr(fp, "name", "<%s>" % fp.__class__.__name__), lineno, 0, line),
                        )

                stack[-1][key] = _unescape(val) if escaped else val

            # exit the loop
            break

    if len(stack) != 1:
        raise SyntaxError(
            "vdf.parse: unclosed parenthasis or quotes (EOF)",
            (getattr(fp, "name", "<%s>" % fp.__class__.__name__), lineno, 0, line),
        )

    return stack.pop()


def loads(s, **kwargs):
    """
    Deserialize ``s`` (a ``str`` or ``unicode`` instance containing a JSON
    document) to a Python object.
    """
    if not isinstance(s, string_type):
        raise TypeError("Expected s to be a str, got %s" % type(s))
    fp = unicodeIO(s)
    return parse(fp, **kwargs)


def load(fp, **kwargs):
    """
    Deserialize ``fp`` (a ``.readline()``-supporting file-like object containing
    a JSON document) to a Python object.
    """
    return parse(fp, **kwargs)


def dumps(obj, pretty=False, escaped=True):
    """
    Serialize ``obj`` to a VDF formatted ``str``.
    """
    if not isinstance(obj, dict):
        raise TypeError("Expected data to be an instance of``dict``")
    if not isinstance(pretty, bool):
        raise TypeError("Expected pretty to be of type bool")
    if not isinstance(escaped, bool):
        raise TypeError("Expected escaped to be of type bool")

    return "".join(_dump_gen(obj, pretty, escaped))


def dump(obj, fp, pretty=False, escaped=True):
    """
    Serialize ``obj`` as a VDF formatted stream to ``fp`` (a
    ``.write()``-supporting file-like object).
    """
    if not isinstance(obj, dict):
        raise TypeError("Expected data to be an instance of``dict``")
    if not hasattr(fp, "write"):
        raise TypeError("Expected fp to have write() method")
    if not isinstance(pretty, bool):
        raise TypeError("Expected pretty to be of type bool")
    if not isinstance(escaped, bool):
        raise TypeError("Expected escaped to be of type bool")

    for chunk in _dump_gen(obj, pretty, escaped):
        fp.write(chunk)


def _dump_gen(data, pretty=False, escaped=True, level=0):
    indent = "\t"
    line_indent = ""

    if pretty:
        line_indent = indent * level

    for key, value in data.items():
        if escaped and isinstance(key, string_type):
            key = _escape(key)

        if isinstance(value, dict):
            yield '%s"%s"\n%s{\n' % (line_indent, key, line_indent)
            for chunk in _dump_gen(value, pretty, escaped, level + 1):
                yield chunk
            yield "%s}\n" % line_indent
        else:
            if escaped and isinstance(value, string_type):
                value = _escape(value)

            yield '%s"%s" "%s"\n' % (line_indent, key, value)


# binary VDF
class BASE_INT(int_type):
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self)


class UINT_64(BASE_INT):
    pass


class INT_64(BASE_INT):
    pass


class POINTER(BASE_INT):
    pass


class COLOR(BASE_INT):
    pass


BIN_NONE = b"\x00"
BIN_STRING = b"\x01"
BIN_INT32 = b"\x02"
BIN_FLOAT32 = b"\x03"
BIN_POINTER = b"\x04"
BIN_WIDESTRING = b"\x05"
BIN_COLOR = b"\x06"
BIN_UINT64 = b"\x07"
BIN_END = b"\x08"
BIN_INT64 = b"\x0a"
BIN_END_ALT = b"\x0b"


def binary_loads(s, mapper=dict, merge_duplicate_keys=True, alt_format=False):
    """
    Deserialize ``s`` (``bytes`` containing a VDF in "binary form")
    to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.
    """
    if not isinstance(s, bytes):
        raise TypeError("Expected s to be bytes, got %s" % type(s))
    if not issubclass(mapper, dict):
        raise TypeError("Expected mapper to be subclass of dict, got %s" % type(mapper))

    # helpers
    int32 = struct.Struct("<i")
    uint64 = struct.Struct("<Q")
    int64 = struct.Struct("<q")
    float32 = struct.Struct("<f")

    def read_string(s, idx, wide=False):
        if wide:
            end = s.find(b"\x00\x00", idx)
            if (end - idx) % 2 != 0:
                end += 1
        else:
            end = s.find(b"\x00", idx)

        if end == -1:
            raise SyntaxError("Unterminated cstring (offset: %d)" % idx)
        result = s[idx:end]
        if wide:
            result = result.decode("utf-16")
        elif bytes is not str:
            result = result.decode("utf-8", "replace")
        else:
            try:
                result.decode("ascii")
            except:
                result = result.decode("utf-8", "replace")
        return result, end + (2 if wide else 1)

    stack = [mapper()]
    idx = 0
    CURRENT_BIN_END = BIN_END if not alt_format else BIN_END_ALT

    while len(s) > idx:
        t = s[idx : idx + 1]
        idx += 1

        if t == CURRENT_BIN_END:
            if len(stack) > 1:
                stack.pop()
                continue
            break

        key, idx = read_string(s, idx)

        if t == BIN_NONE:
            if merge_duplicate_keys and key in stack[-1]:
                _m = stack[-1][key]
            else:
                _m = mapper()
                stack[-1][key] = _m
            stack.append(_m)
        elif t == BIN_STRING:
            stack[-1][key], idx = read_string(s, idx)
        elif t == BIN_WIDESTRING:
            stack[-1][key], idx = read_string(s, idx, wide=True)
        elif t in (BIN_INT32, BIN_POINTER, BIN_COLOR):
            val = int32.unpack_from(s, idx)[0]

            if t == BIN_POINTER:
                val = POINTER(val)
            elif t == BIN_COLOR:
                val = COLOR(val)

            stack[-1][key] = val
            idx += int32.size
        elif t == BIN_UINT64:
            stack[-1][key] = UINT_64(uint64.unpack_from(s, idx)[0])
            idx += uint64.size
        elif t == BIN_INT64:
            stack[-1][key] = INT_64(int64.unpack_from(s, idx)[0])
            idx += int64.size
        elif t == BIN_FLOAT32:
            stack[-1][key] = float32.unpack_from(s, idx)[0]
            idx += float32.size
        else:
            raise SyntaxError("Unknown data type at offset %d: %s" % (idx - 1, repr(t)))

    if len(s) != idx or len(stack) != 1:
        raise SyntaxError("Binary VDF ended at offset %d, but length is %d" % (idx, len(s)))

    return stack.pop()


def binary_dumps(obj, alt_format=False):
    """
    Serialize ``obj`` to a binary VDF formatted ``bytes``.
    """
    return b"".join(_binary_dump_gen(obj, alt_format=alt_format))


def _binary_dump_gen(obj, level=0, alt_format=False):
    if level == 0 and len(obj) == 0:
        return

    int32 = struct.Struct("<i")
    uint64 = struct.Struct("<Q")
    int64 = struct.Struct("<q")
    float32 = struct.Struct("<f")

    for key, value in obj.items():
        if isinstance(key, string_type):
            key = key.encode("utf-8")
        else:
            raise TypeError("dict keys must be of type str, got %s" % type(key))

        if isinstance(value, dict):
            yield BIN_NONE + key + BIN_NONE
            for chunk in _binary_dump_gen(value, level + 1, alt_format=alt_format):
                yield chunk
        elif isinstance(value, UINT_64):
            yield BIN_UINT64 + key + BIN_NONE + uint64.pack(value)
        elif isinstance(value, INT_64):
            yield BIN_INT64 + key + BIN_NONE + int64.pack(value)
        elif isinstance(value, string_type):
            try:
                value = value.encode("utf-8") + BIN_NONE
                yield BIN_STRING
            except:
                value = value.encode("utf-16") + BIN_NONE * 2
                yield BIN_WIDESTRING
            yield key + BIN_NONE + value
        elif isinstance(value, float):
            yield BIN_FLOAT32 + key + BIN_NONE + float32.pack(value)
        elif isinstance(value, (COLOR, POINTER, int, int_type)):
            if isinstance(value, COLOR):
                yield BIN_COLOR
            elif isinstance(value, POINTER):
                yield BIN_POINTER
            else:
                yield BIN_INT32
            yield key + BIN_NONE
            yield int32.pack(value)
        else:
            raise TypeError("Unsupported type: %s" % type(value))

    yield BIN_END if not alt_format else BIN_END_ALT


def vbkv_loads(s, mapper=dict, merge_duplicate_keys=True):
    """
    Deserialize ``s`` (``bytes`` containing a VBKV to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.
    """
    if s[:4] != b"VBKV":
        raise ValueError("Invalid header")

    (checksum,) = struct.unpack("<i", s[4:8])

    if checksum != crc32(s[8:]):
        raise ValueError("Invalid checksum")

    return binary_loads(s[8:], mapper, merge_duplicate_keys, alt_format=True)


def vbkv_dumps(obj):
    """
    Serialize ``obj`` to a VBKV formatted ``bytes``.
    """
    data = b"".join(_binary_dump_gen(obj, alt_format=True))
    checksum = crc32(data)

    return b"VBKV" + struct.pack("<i", checksum) + data



# Standard Library
import os
import subprocess
import xml.etree.ElementTree as etree
from gettext import gettext as _

# Lutris Modules
from lutris import settings
from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.log import logger

SNES9X_DIR = os.path.join(settings.DATA_DIR, "runners/snes9x")


class snes9x(Runner):
    description = _("Super Nintendo emulator")
    human_name = _("Snes9x")
    platforms = [_("Nintendo SNES")]
    runnable_alone = True
    runner_executable = "snes9x/bin/snes9x-gtk"
    flatpak_id = "com.snes9x.Snes9x"
    use_sniper_runtime = True
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "default_path": "game_path",
            "label": _("ROM file"),
            "help": _("The game data, commonly called a ROM image."),
        }
    ]

    runner_options = [
        {"option": "fullscreen", "type": "bool", "section": _("Graphics"), "label": _("Fullscreen"), "default": "1"},
        {
            "option": "maintain_aspect_ratio",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Maintain aspect ratio (4:3)"),
            "default": "1",
            "help": _(
                "Super Nintendo games were made for 4:3 "
                "screens with rectangular pixels, but modern screens "
                "have square pixels, which results in a vertically "
                "squeezed image. This option corrects this by displaying "
                "rectangular pixels."
            ),
        },
        {
            "option": "sound_driver",
            "type": "choice",
            "label": _("Sound driver"),
            "advanced": True,
            "choices": (("SDL", "1"), ("ALSA", "2"), ("OSS", "0")),
            "default": "1",
        },
    ]

    def set_option(self, option, value):
        config_file = os.path.expanduser("~/.snes9x/snes9x.xml")
        if not system.path_exists(config_file):
            with subprocess.Popen(self.get_command() + ["-help"]) as snes9x_process:
                snes9x_process.communicate()
        if not system.path_exists(config_file):
            logger.error("Snes9x config file creation failed")
            return
        tree = etree.parse(config_file)
        node = tree.find("./preferences/option[@name='%s']" % option)
        if node is not None:  # some settings are just for Lutris, they won't be in the file
            if isinstance(value, bool):
                value = "1" if value else "0"
            node.attrib["value"] = value
            tree.write(config_file)

    def play(self):
        for option_name in self.config.runner_config:
            self.set_option(option_name, self.runner_config.get(option_name))

        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        return {"command": self.get_command() + [rom]}



"""Service package"""

import os

from lutris import settings
from lutris.services.amazon import AmazonService
from lutris.services.battlenet import BNET_ENABLED, BattleNetService
from lutris.services.dolphin import DolphinService
from lutris.services.ea_app import EAAppService
from lutris.services.egs import EpicGamesStoreService
from lutris.services.flathub import FlathubService
from lutris.services.gog import GOGService
from lutris.services.humblebundle import HumbleBundleService
from lutris.services.itchio import ItchIoService
from lutris.services.lutris import LutrisService
from lutris.services.mame import MAMEService
from lutris.services.scummvm import SCUMMVM_CONFIG_FILE, ScummvmService
from lutris.services.steam import SteamService
from lutris.services.steamfamily import SteamFamilyService
from lutris.services.steamwindows import SteamWindowsService
from lutris.services.ubisoft import UbisoftConnectService
from lutris.services.xdg import XDGService
from lutris.services.zoom import ZoomService
from lutris.util import system
from lutris.util.dolphin.cache_reader import DOLPHIN_GAME_CACHE_FILE
from lutris.util.linux import LINUX_SYSTEM

DEFAULT_SERVICES = ["gog", "egs", "ea_app", "ubisoft", "steam"]


def get_services():
    """Return a mapping of available services"""
    _services = {
        "gog": GOGService,
        "zoomplatform": ZoomService,
        "humblebundle": HumbleBundleService,
        "egs": EpicGamesStoreService,
        "itchio": ItchIoService,
        "ea_app": EAAppService,
        "ubisoft": UbisoftConnectService,
        "amazon": AmazonService,
        "flathub": FlathubService,
    }
    if BNET_ENABLED:
        _services["battlenet"] = BattleNetService
    if not LINUX_SYSTEM.is_flatpak():
        _services["xdg"] = XDGService
    if LINUX_SYSTEM.has_steam():
        _services["steam"] = SteamService
    _services["steamwindows"] = SteamWindowsService
    _services["steamfamily"] = SteamFamilyService
    if system.path_exists(DOLPHIN_GAME_CACHE_FILE):
        _services["dolphin"] = DolphinService
    if system.path_exists(SCUMMVM_CONFIG_FILE):
        _services["scummvm"] = ScummvmService
    if os.environ.get("LUTRIS_SERVICE_ENABLED") == "1":
        _services["lutris"] = LutrisService
    return _services


SERVICES = get_services()


# Those services are not yet ready to be used
WIP_SERVICES = {
    "mame": MAMEService,
}

if os.environ.get("LUTRIS_ENABLE_ALL_SERVICES"):
    SERVICES.update(WIP_SERVICES)


def get_enabled_services():
    return {
        key: _class
        for key, _class in SERVICES.items()
        if settings.read_setting(key, section="services").lower() == "true"
    }



import bisect
import os
import threading
import time
from typing import Any, Dict, Optional

import requests

from lutris import __version__
from lutris.util import jobs
from lutris.util.log import logger

# `time.time` can skip ahead or even go backwards if the current
# system time is changed between invocations. Use `time.monotonic`
# so we won't have screenshots making fun of us for showing negative
# download speeds.
get_time = time.monotonic


class Downloader:
    """Non-blocking downloader.

    Do start() then check_progress() at regular intervals.
    Download is done when check_progress() returns 1.0.
    Stop with cancel().
    """

    (INIT, DOWNLOADING, CANCELLED, ERROR, COMPLETED) = list(range(5))

    def __init__(
        self,
        url: str,
        dest: str,
        overwrite: bool = False,
        referer: Optional[str] = None,
        cookies: Any = None,
        headers: Dict[str, str] = None,
    ) -> None:
        self.url: str = url
        self.dest: str = dest
        self.cookies = cookies
        self.headers = headers
        self.overwrite: bool = overwrite
        self.referer = referer
        self.stop_request = None
        self.thread = None

        # Read these after a check_progress()
        self.state = self.INIT
        self.error = None
        self.downloaded_size: int = 0  # Bytes
        self.full_size: int = 0  # Bytes
        self.progress_fraction: float = 0
        self.progress_percentage: float = 0
        self.average_speed = 0
        self.time_left: str = "00:00:00"  # Based on average speed
        self.last_size: int = 0
        self.last_check_time: float = 0.0
        self.last_speeds = []
        self.speed_check_time = 0
        self.time_left_check_time = 0
        self.file_pointer = None
        self.progress_event = threading.Event()

    def __repr__(self):
        return "downloader for %s" % self.url

    def start(self):
        """Start download job."""
        logger.debug("⬇ %s", self.url)
        self.state = self.DOWNLOADING
        self.last_check_time = get_time()
        if self.overwrite and os.path.isfile(self.dest):
            os.remove(self.dest)
        self.file_pointer = open(self.dest, "wb")  # pylint: disable=consider-using-with
        self.thread = jobs.AsyncCall(self.async_download, None)
        self.stop_request = self.thread.stop_request

    def reset(self):
        """Reset the state of the downloader"""
        self.state = self.INIT
        self.error = None
        self.downloaded_size = 0  # Bytes
        self.full_size = 0  # Bytes
        self.progress_fraction = 0
        self.progress_percentage = 0
        self.average_speed = 0
        self.time_left = "00:00:00"  # Based on average speed
        self.last_size = 0
        self.last_check_time = 0.0
        self.last_speeds = []
        self.speed_check_time = 0
        self.time_left_check_time = 0
        self.file_pointer = None

    def check_progress(self, blocking=False):
        """Append last downloaded chunk to dest file and store stats.

        blocking: if true and still downloading, block until some progress is made.
        :return: progress (between 0.0 and 1.0)"""
        if blocking and self.state in [self.INIT, self.DOWNLOADING] and self.progress_fraction < 1.0:
            self.progress_event.wait()
            self.progress_event.clear()

        if self.state not in [self.CANCELLED, self.ERROR]:
            self.get_stats()
        return self.progress_fraction

    def join(self, progress_callback=None):
        """Blocks waiting for the download to complete.

        'progress_callback' is invoked repeatedly as the download
        proceeds, if given, and is passed the downloader itself.

        Returns True on success, False if cancelled."""
        while self.state == self.DOWNLOADING:
            self.check_progress(blocking=True)
            if progress_callback:
                progress_callback(self)

        if self.error:
            raise self.error
        return self.state == self.COMPLETED

    def cancel(self):
        """Request download stop and remove destination file."""
        logger.debug("❌ %s", self.url)
        self.state = self.CANCELLED
        if self.stop_request:
            self.stop_request.set()
        if self.file_pointer:
            self.file_pointer.close()
            self.file_pointer = None
        if os.path.isfile(self.dest):
            os.remove(self.dest)

    def async_download(self):
        try:
            headers = requests.utils.default_headers()
            headers["User-Agent"] = "Lutris/%s" % __version__
            if self.referer:
                headers["Referer"] = self.referer
            if self.headers:
                for key, value in self.headers.items():
                    headers[key] = value
            response = requests.get(self.url, headers=headers, stream=True, timeout=30, cookies=self.cookies)
            if response.status_code != 200:
                logger.info("%s returned a %s error", self.url, response.status_code)
            response.raise_for_status()
            self.full_size = int(response.headers.get("Content-Length", "").strip() or 0)
            self.progress_event.set()
            for chunk in response.iter_content(chunk_size=8192):
                if not self.file_pointer:
                    break
                if chunk:
                    self.downloaded_size += len(chunk)
                    self.file_pointer.write(chunk)
                self.progress_event.set()
            self.on_download_completed()
        except Exception as ex:
            logger.exception("Download failed: %s", ex)
            self.on_download_failed(ex)

    def on_download_failed(self, error: Exception):
        # Cancelling closes the file, which can result in an
        # error. If so, we just remain cancelled.
        if self.state != self.CANCELLED:
            self.state = self.ERROR
            self.error = error
        if self.file_pointer:
            self.file_pointer.close()
            self.file_pointer = None

    def on_download_completed(self):
        if self.state == self.CANCELLED:
            return

        logger.debug("Finished downloading %s", self.url)
        if not self.downloaded_size:
            logger.warning("Downloaded file is empty")

        if not self.full_size:
            self.progress_fraction = 1.0
            self.progress_percentage = 100
        self.state = self.COMPLETED
        self.file_pointer.close()
        self.file_pointer = None

    def get_stats(self):
        """Calculate and store download stats."""
        self.average_speed = self.get_speed()
        self.time_left = self.get_average_time_left()
        self.last_check_time = get_time()
        self.last_size = self.downloaded_size

        if self.full_size:
            self.progress_fraction = float(self.downloaded_size) / float(self.full_size)
            self.progress_percentage = self.progress_fraction * 100

    def get_speed(self):
        """Return the average speed of the download so far."""
        elapsed_time = get_time() - self.last_check_time
        if elapsed_time > 0:
            chunk_size = self.downloaded_size - self.last_size
            speed = chunk_size / elapsed_time or 1
            # insert in sorted order, so we can omit the least and
            # greatest value later
            bisect.insort(self.last_speeds, speed)

        # Until we get the first sample, just return our default
        if not self.last_speeds:
            return self.average_speed

        if get_time() - self.speed_check_time < 1:  # Minimum delay
            return self.average_speed

        while len(self.last_speeds) > 20:
            self.last_speeds.pop(0)

        if len(self.last_speeds) > 7:
            # Skip extreme values
            samples = self.last_speeds[1:-1]
        else:
            samples = self.last_speeds[:]

        average_speed = sum(samples) / len(samples)

        self.speed_check_time = get_time()
        return average_speed

    def get_average_time_left(self) -> str:
        """Return average download time left as string."""
        if not self.full_size:
            return "???"

        elapsed_time = get_time() - self.time_left_check_time
        if elapsed_time < 1:  # Minimum delay
            return self.time_left

        average_time_left = (self.full_size - self.downloaded_size) / self.average_speed
        minutes, seconds = divmod(average_time_left, 60)
        hours, minutes = divmod(minutes, 60)
        self.time_left_check_time = get_time()
        return "%d:%02d:%02d" % (hours, minutes, seconds)



"""XrandR based display management"""

import re
import subprocess
from collections import namedtuple

from lutris.settings import DEFAULT_RESOLUTION_HEIGHT, DEFAULT_RESOLUTION_WIDTH
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.system import read_process_output

Output = namedtuple("Output", ("name", "mode", "position", "rotation", "primary", "rate"))


def _get_vidmodes():
    """Return video modes from XrandR"""
    xrandr_output = read_process_output([LINUX_SYSTEM.get("xrandr")]).split("\n")
    logger.debug("Retrieving %s video modes from XrandR", len(xrandr_output))
    return xrandr_output


def get_outputs():  # pylint: disable=too-many-locals
    """Return list of namedtuples containing output 'name', 'geometry',
    'rotation' and whether it is the 'primary' display."""
    outputs = []
    logger.debug("Retrieving display outputs")
    vid_modes = _get_vidmodes()
    position = None
    rotate = None
    primary = None
    name = None
    if not vid_modes:
        logger.error("xrandr didn't return anything")
        return []
    for line in vid_modes:
        fields = line.split()
        if "connected" in fields[1:] and len(fields) >= 4:
            try:
                connected_index = fields.index("connected", 1)
                name_fields = fields[:connected_index]
                name = " ".join(name_fields)
                data_fields = fields[connected_index + 1 :]
                if data_fields[0] == "primary":
                    data_fields = data_fields[1:]
                geometry, rotate, *_ = data_fields
                if geometry.startswith("("):  # Screen turned off, no geometry
                    continue
                if rotate.startswith("("):  # Screen not rotated, no need to include
                    rotate = "normal"
                _, x_pos, y_pos = geometry.split("+")
                position = "{x_pos}x{y_pos}".format(x_pos=x_pos, y_pos=y_pos)
            except ValueError as ex:
                logger.error(
                    "Unhandled xrandr line %s, error: %s. Please send your xrandr output to the dev team", line, ex
                )
                continue
        elif "*" in line:
            mode, *framerates = fields
            for number in framerates:
                if "*" in number:
                    hertz = number[:-2]
                    outputs.append(
                        Output(
                            name=name,
                            mode=mode,
                            position=position,
                            rotation=rotate,
                            primary=primary,
                            rate=hertz,
                        )
                    )
                    break
    return outputs


def turn_off_except(display):
    """Use XrandR to turn off displays except the one referenced by `display`"""
    if not display:
        logger.error("No active display given, no turning off every display")
        return
    for output in get_outputs():
        if output.name != display:
            logger.info("Turning off %s", output[0])
            with subprocess.Popen([LINUX_SYSTEM.get("xrandr"), "--output", output.name, "--off"]) as xrandr:
                xrandr.communicate()


def get_resolutions():
    """Return the list of supported screen resolutions."""
    resolution_list = []
    logger.debug("Retrieving resolution list")
    for line in _get_vidmodes():
        if line.startswith("  "):
            resolution_match = re.match(r".*?(\d+x\d+).*", line)
            if resolution_match:
                resolution_list.append(resolution_match.groups()[0])
    if not resolution_list:
        logger.error("Unable to generate resolution list from xrandr output")
        return ["%sx%s" % (DEFAULT_RESOLUTION_WIDTH, DEFAULT_RESOLUTION_HEIGHT)]
    return sorted(set(resolution_list), key=lambda x: int(x.split("x")[0]), reverse=True)


def change_resolution(resolution):
    """Change display resolution.

    Takes a string for single monitors or a list of displays as returned
    by get_outputs().
    """
    if not resolution:
        logger.warning("No resolution provided")
        return
    if isinstance(resolution, str):
        logger.debug("Switching resolution to %s", resolution)

        if resolution not in get_resolutions():
            logger.warning("Resolution %s doesn't exist.", resolution)
        else:
            output_name = get_outputs()[0].name
            logger.info("Changing resolution on %s to %s", output_name, resolution)
            args = [LINUX_SYSTEM.get("xrandr"), "--output", output_name, "--mode", resolution]
            with subprocess.Popen(args) as xrandr:
                xrandr.communicate()
    else:
        for display in resolution:
            logger.debug("Switching to %s on %s", display.mode, display.name)

            if display.rotation is not None and display.rotation in (
                "normal",
                "left",
                "right",
                "inverted",
            ):
                rotation = display.rotation
            else:
                rotation = "normal"
            logger.info("Switching resolution of %s to %s", display.name, display.mode)
            with subprocess.Popen(
                [
                    LINUX_SYSTEM.get("xrandr"),
                    "--output",
                    display.name,
                    "--mode",
                    display.mode,
                    "--pos",
                    display.position,
                    "--rotate",
                    rotation,
                    "--rate",
                    display.rate,
                ]
            ) as xrandr:
                xrandr.communicate()


class LegacyDisplayManager:  # pylint: disable=too-few-public-methods
    """Legacy XrandR based display manager.
    Does not work on Wayland.
    """

    @staticmethod
    def get_display_names():
        """Return output names from XrandR"""
        return [output.name for output in get_outputs()]

    @staticmethod
    def get_resolutions():
        """Return available resolutions"""
        return get_resolutions()

    @staticmethod
    def get_current_resolution():
        """Return the current resolution for the desktop"""
        for line in _get_vidmodes():
            if line.startswith("  ") and "*" in line:
                resolution_match = re.match(r".*?(\d+x\d+).*", line)
                if resolution_match:
                    return resolution_match.groups()[0].split("x")
        logger.error("Unable to find the current resolution from xrandr output")
        return str(DEFAULT_RESOLUTION_WIDTH), str(DEFAULT_RESOLUTION_HEIGHT)

    @staticmethod
    def set_resolution(resolution):
        """Change the current resolution"""
        change_resolution(resolution)

    @staticmethod
    def get_config():
        """Return the current display configuration"""
        return get_outputs()



"""Manipulates installer files"""

from gettext import gettext as _
from urllib.parse import urlparse

from lutris.cache import has_valid_custom_cache_path
from lutris.gui.widgets.download_collection_progress_box import DownloadCollectionProgressBox
from lutris.util import system
from lutris.util.strings import gtk_safe_urls

AMAZON_DOMAIN = "a2z.com"


class InstallerFileCollection:
    """Representation of a collection of files in the `files` sections of an installer.
    Store files in a folder"""

    def __init__(self, game_slug, file_id, files_list):
        self.game_slug = game_slug
        self.id = file_id.replace("-", "_")  # pylint: disable=invalid-name
        self.num_files = len(files_list)
        self.files_list = files_list
        self._dest_file_override = None  # Used to override the destination
        self.full_size = 0
        self._get_files_size()
        self._get_service()

    def _get_files_size(self):
        if len(self.files_list) > 0:
            if self.files_list[0].total_size:
                self.full_size = self.files_list[0].total_size
            else:
                self.full_size = sum(f.size or 0 for f in self.files_list)

    def _get_service(self):
        """Try to get the service using the url of an InstallerFile"""
        self.service = None
        if len(self.files_list) < 1:
            return
        url = self.files_list[0].url
        url_parts = urlparse(url)
        if url_parts.netloc.endswith(AMAZON_DOMAIN):
            self.service = "amazon"

    def copy(self):
        """Copy InstallerFileCollection"""
        # copy all InstallerFile inside file list
        new_file_list = []
        for file in self.files_list:
            new_file_list.append(file.copy())
        collection = InstallerFileCollection(self.game_slug, self.id, new_file_list)
        collection._dest_file_override = self._dest_file_override
        return collection

    def override_dest_file(self, new_dest_file):
        """Called by the UI when the user selects a file path; this causes
        the collection to be ready if this one file is there, and
        we'll special case GOG here too."""
        self._dest_file_override = new_dest_file

        if len(self.files_list) == 1:
            self.files_list[0].override_dest_file(new_dest_file)
        else:
            # try to set main gog file to dest_file
            for installer_file in self.files_list:
                if installer_file.id == "goginstaller":
                    installer_file.dest_file = new_dest_file

    @property
    def is_dest_file_overridden(self):
        return bool(self._dest_file_override)

    def get_dest_files_by_id(self):
        files = {}
        for file in self.files_list:
            files.update({file.id: file.dest_file})
        return files

    def __str__(self):
        return "%s/%s" % (self.game_slug, self.id)

    @property
    def auxiliary_info(self):
        """Provides a small bit of additional descriptive texts to show in the UI."""
        if self.num_files == 1:
            return f"{self.num_files} {_('File')}"

        return f"{self.num_files} {_('Files')}"

    @property
    def human_url(self):
        """Return game_slug"""
        return self.game_slug

    def get_label(self):
        """Return a human readable label for installer files"""
        return gtk_safe_urls(self.game_slug)

    @property
    def default_provider(self):
        """Return file provider used. File Collection only supports 'pga' and 'download'"""
        if self.is_cached:
            return "pga"
        return "download"

    @property
    def providers(self):
        """Return all supported providers. File Collection only supports 'pga' and 'download'"""
        _providers = set()
        if self.is_cached:
            _providers.add("pga")
        _providers.add("download")
        return _providers

    def uses_pga_cache(self):
        """Determines whether the installer files are stored in a PGA cache

        Returns:
            bool
        """
        return has_valid_custom_cache_path()

    @property
    def is_user_pga_caching_allowed(self):
        return len(self.files_list) == 1 and self.files_list[0].is_user_pga_caching_allowed

    def prepare(self):
        """Prepare the file for download, if we've not been redirected to an existing file."""
        if not self.is_dest_file_overridden or len(self.files_list) == 1:
            for installer_file in self.files_list:
                installer_file.prepare()

    def create_download_progress_box(self):
        return DownloadCollectionProgressBox(self)

    def is_ready(self, provider):
        """Is the file already present at the destination (if applicable)?"""
        if provider not in ("user", "pga"):
            return True

        if self._dest_file_override:
            return system.path_exists(self._dest_file_override)

        for installer_file in self.files_list:
            if not installer_file.is_ready(provider):
                return False
        return True

    def check_hash(self):
        """Check the hash of all the files (if available)"""
        for installer_file in self.files_list:
            installer_file.check_hash()

    @property
    def is_cached(self):
        """Are the files available in the local PGA cache?"""
        if self.uses_pga_cache():
            # check if every file is in cache, without checking
            # uses_pga_cache() on each.
            for installer_file in self.files_list:
                if not system.path_exists(installer_file.dest_file):
                    return False
            return True
        return False

    def save_to_cache(self):
        """Copy the files into the PGA cache."""
        for installer_file in self.files_list:
            installer_file.save_to_cache()

    def remove_previous(self):
        """Remove file already at destination, prior to starting the download."""
        for installer_file in self.files_list:
            installer_file.remove_previous()



"""Language and translation utilities"""

import locale

from lutris.config import LutrisConfig
from lutris.util.log import logger


def get_user_locale():
    """Get locale for the user, based on global options, else try system locale"""
    config = LutrisConfig(level="system")
    if config.system_config.get("locale"):
        try:
            user_locale, _user_encoding = config.system_config["locale"].split(".")
            return user_locale
        except ValueError:  # If '.' is not found
            return config.system_config["locale"]

    user_locale, _user_encoding = locale.getlocale()
    if not user_locale:
        logger.error("Unable to get locale")
        return
    return user_locale


def get_lang():
    """Return the 2 letter language code used by the system"""
    user_locale = get_user_locale()
    if not user_locale:
        return ""
    return user_locale[:2]


def get_lang_and_country():
    """Return language code and country for the current user"""
    user_locale = get_user_locale()
    if not user_locale:
        return "", ""
    lang_code, country = user_locale.split("-" if "-" in locale else "_")
    return lang_code, country



from gettext import gettext as _

from gi.repository import GObject, Gtk

from lutris import runners
from lutris.gui.config.runner import RunnerConfigDialog
from lutris.gui.dialogs import QuestionDialog
from lutris.gui.dialogs.runner_install import RunnerInstallDialog
from lutris.gui.widgets.scaled_image import ScaledImage
from lutris.util.log import logger


class RunnerBox(Gtk.Box):
    __gsignals__ = {
        "runner-installed": (GObject.SIGNAL_RUN_FIRST, None, ()),
        "runner-removed": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, runner_name):
        super().__init__(visible=True)

        self.connect("runner-installed", self.on_runner_installed)
        self.connect("runner-removed", self.on_runner_removed)

        self.set_margin_bottom(12)
        self.set_margin_top(12)
        self.set_margin_left(12)
        self.set_margin_right(12)
        self.runner = runners.import_runner(runner_name)()

        runner_icon = ScaledImage.get_runtime_icon_image(
            self.runner.name, scale_factor=self.get_scale_factor(), visible=True
        )
        runner_icon.set_margin_right(12)
        self.pack_start(runner_icon, False, True, 6)

        self.runner_label_box = Gtk.VBox(visible=True)
        self.runner_label_box.set_margin_top(12)

        runner_label = Gtk.Label(visible=True)
        runner_label.set_alignment(0, 0.5)
        runner_label.set_markup("<b>%s</b>" % self.runner.human_name)
        self.runner_label_box.pack_start(runner_label, False, False, 0)

        desc_label = Gtk.Label(visible=True)
        desc_label.set_line_wrap(True)
        desc_label.set_alignment(0, 0.5)
        desc_label.set_text(self.runner.description)
        self.runner_label_box.pack_start(desc_label, False, False, 0)

        self.pack_start(self.runner_label_box, True, True, 0)

        self.configure_button = Gtk.Button.new_from_icon_name("preferences-system-symbolic", Gtk.IconSize.BUTTON)
        self.configure_button.set_valign(Gtk.Align.CENTER)
        self.configure_button.set_margin_right(12)
        self.configure_button.connect("clicked", self.on_configure_clicked)
        self.pack_start(self.configure_button, False, False, 0)
        if not self.runner.is_installed():
            self.runner_label_box.set_sensitive(False)
        self.configure_button.show()
        self.action_alignment = Gtk.Alignment.new(0.5, 0.5, 0, 0)
        self.action_alignment.show()
        self.action_alignment.add(self.get_action_button())
        self.pack_start(self.action_alignment, False, False, 0)

    def get_action_button(self):
        """Return a install or remove button"""
        if self.runner.multiple_versions:
            _button = Gtk.Button.new_from_icon_name("system-software-install-symbolic", Gtk.IconSize.BUTTON)
            _button.get_style_context().add_class("circular")
            _button.connect("clicked", self.on_versions_clicked)
        else:
            if self.runner.can_uninstall():
                _button = Gtk.Button.new_from_icon_name("edit-delete-symbolic", Gtk.IconSize.BUTTON)
                _button.get_style_context().add_class("circular")
                _button.connect("clicked", self.on_remove_clicked)
                _button.set_sensitive(self.runner.can_uninstall())
            else:
                _button = Gtk.Button.new_from_icon_name("system-software-install-symbolic", Gtk.IconSize.BUTTON)
                _button.get_style_context().add_class("circular")
                _button.connect("clicked", self.on_install_clicked)
                _button.set_sensitive(not self.runner.is_installed(flatpak_allowed=False))
        _button.show()
        return _button

    def on_versions_clicked(self, widget):
        window = self.get_toplevel()
        application = window.get_application()
        title = _("Manage %s versions") % self.runner.name
        application.show_window(RunnerInstallDialog, title=title, runner=self.runner, parent=window)

    def on_install_clicked(self, widget):
        """Install a runner."""
        logger.debug("Install of %s requested", self.runner)
        self.runner.install(self.get_toplevel())

        if self.runner.is_installed():
            self.emit("runner-installed")

    def on_configure_clicked(self, widget):
        window = self.get_toplevel()
        application = window.get_application()
        application.show_window(RunnerConfigDialog, runner=self.runner, parent=window)

    def on_remove_clicked(self, widget):
        dialog = QuestionDialog(
            {
                "parent": self.get_toplevel(),
                "title": _("Do you want to uninstall %s?") % self.runner.human_name,
                "question": _("This will remove <b>%s</b> and all associated data." % self.runner.human_name),
            }
        )
        if Gtk.ResponseType.YES == dialog.result:

            def on_runner_uninstalled():
                self.emit("runner-removed")

            self.runner.uninstall(on_runner_uninstalled)

    def on_runner_installed(self, widget):
        """Called after the runnner is installed"""
        self.runner_label_box.set_sensitive(True)
        self.action_alignment.get_children()[0].destroy()
        self.action_alignment.add(self.get_action_button())

    def on_runner_removed(self, widget):
        """Called after the runner is removed"""
        self.runner_label_box.set_sensitive(False)
        self.action_alignment.get_children()[0].destroy()
        self.action_alignment.add(self.get_action_button())



"""Base module for runners"""

import os
import signal
from gettext import gettext as _
from typing import Any, Callable, Dict, Iterable, Optional, Set

from lutris import runtime, settings
from lutris.api import format_runner_version, get_default_runner_version_info
from lutris.config import LutrisConfig
from lutris.database.games import get_game_by_field
from lutris.exceptions import MisconfigurationError, MissingExecutableError, UnavailableLibrariesError
from lutris.monitored_command import MonitoredCommand
from lutris.runners import RunnerInstallationError
from lutris.util import flatpak, strings, system
from lutris.util.extract import ExtractError, extract_archive
from lutris.util.graphics.gpu import GPUS
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.process import Process
from lutris.util.sniper import get_sniper_ld_library_path, get_sniper_run_command


def kill_processes(sig: int, pids: Iterable[int]) -> None:
    """Sends a signal to a process list, logging errors without stopping."""
    for pid in pids:
        try:
            os.kill(int(pid), sig)
        except ProcessLookupError as ex:
            logger.debug("Failed to kill game process: %s", ex)
        except PermissionError:
            logger.debug("Permission to kill process %s denied", pid)


class Runner:  # pylint: disable=too-many-public-methods
    """Generic runner (base class for other runners)."""

    multiple_versions = False
    platforms = []
    runnable_alone = False
    game_options = []
    runner_options = []
    system_options_override = []
    context_menu_entries = []
    require_libs = []
    runner_executable = None
    entry_point_option = "main_file"
    download_url = None
    arch = None  # If the runner is only available for an architecture that isn't x86_64
    flatpak_id = None
    human_name = ""
    use_sniper_runtime = False

    def __init__(self, config=None):
        """Initialize runner."""
        if config:
            self.has_explicit_config = True
            self._config = config
            self.game_data = get_game_by_field(config.game_config_id, "configpath")
        else:
            self.has_explicit_config = False
            self._config = None
            self.game_data = {}

    def __lt__(self, other):
        return self.name < other.name

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def directory(self):
        return os.path.join(settings.RUNNER_DIR, self.name)

    @property
    def config(self):
        if not self._config:
            self._config = LutrisConfig(runner_slug=self.name)
        return self._config

    @config.setter
    def config(self, new_config):
        self._config = new_config
        self.has_explicit_config = new_config is not None

    @property
    def game_config(self):
        """Return the cascaded game config as a dict."""
        return self.config.game_config

    @property
    def runner_config(self):
        """Return the cascaded runner config as a dict."""
        return self.config.runner_config

    @property
    def system_config(self):
        """Return the cascaded system config as a dict."""
        return self.config.system_config

    @property
    def default_path(self):
        """Return the default path where games are installed."""
        return self.system_config.get("game_path")

    @property
    def game_path(self):
        """Return the directory where the game is installed."""
        game_path = self.game_data.get("directory")
        if game_path:
            return os.path.expanduser(game_path)  # expanduser just in case!

        if self.has_explicit_config:
            # Default to the directory where the entry point is located.
            entry_point = self.game_config.get(self.entry_point_option)
            if entry_point:
                return os.path.dirname(os.path.expanduser(entry_point))
        return ""

    def resolve_game_path(self):
        """Returns the path where the game is found; if game_path does not
        provide a path, this may try to resolve the path by runner-specific means,
        which can find things like /usr/games when applicable."""
        return self.game_path

    @property
    def working_dir(self):
        """Return the working directory to use when running the game."""
        return self.game_path or os.path.expanduser("~/")

    @property
    def shader_cache_dir(self):
        """Return the cache directory for this runner to use. We create
        this if it does not exist."""
        path = os.path.join(settings.SHADER_CACHE_DIR, self.name)
        if not os.path.isdir(path):
            os.mkdir(path)
        return path

    @property
    def nvidia_shader_cache_path(self):
        """The path to place in __GL_SHADER_DISK_CACHE_PATH; NVidia
        will place its cache cache in a subdirectory here."""
        return self.shader_cache_dir

    @property
    def discord_client_id(self):
        if self.game_data.get("discord_client_id"):
            return self.game_data.get("discord_client_id")

    def get_platform(self):
        return self.platforms[0]

    def get_runner_options(self):
        runner_options = self.runner_options[:]
        if self.runner_executable:
            runner_options.append(
                {
                    "option": "runner_executable",
                    "type": "file",
                    "label": _("Custom executable for the runner"),
                    "advanced": True,
                }
            )

        runner_options.append(
            {
                "section": _("Side Panel"),
                "option": "visible_in_side_panel",
                "type": "bool",
                "label": _("Visible in Side Panel"),
                "default": True,
                "advanced": True,
                "scope": ["runner"],
                "help": _("Show this runner in the side panel if it is installed or available through Flatpak."),
            }
        )
        return runner_options

    def get_executable(self) -> str:
        if "runner_executable" in self.runner_config:
            runner_executable = self.runner_config["runner_executable"]
            if os.path.isfile(runner_executable):
                return runner_executable
        if not self.runner_executable:
            raise MisconfigurationError("runner_executable not set for {}".format(self.name))

        exe = os.path.join(settings.RUNNER_DIR, self.runner_executable)
        if not os.path.isfile(exe):
            raise MissingExecutableError(_("The executable '%s' could not be found.") % exe)
        return exe

    def get_command(self):
        """Returns the command line to run the runner itself; generally a game
        will be appended to this by play()."""
        try:
            exe = self.get_executable()
            if not system.path_exists(exe):
                raise MissingExecutableError(_("The executable '%s' could not be found.") % exe)
            command = [exe]
        except MisconfigurationError:
            if flatpak.is_app_installed(self.flatpak_id):
                return flatpak.get_run_command(self.flatpak_id)

            raise

        if self.use_sniper_runtime:
            sniper_cmd = get_sniper_run_command()
            if sniper_cmd:
                # pressure-vessel overrides LD_LIBRARY_PATH, so we must set
                # library paths inside the container via bash -c. Sniper runtime
                # paths come first (so module-loading libs like gdk-pixbuf use
                # their own loaders), then host /run/host/ paths as fallback for
                # libraries missing from Sniper (e.g. libgtkmm-3.0, libpulsecommon).
                host_lib_paths = get_sniper_ld_library_path()
                shell_cmd = 'export LD_LIBRARY_PATH="$LD_LIBRARY_PATH:%s"; exec "$@"' % host_lib_paths
                command = [sniper_cmd, "--", "bash", "-c", shell_cmd, "bash"] + command
            else:
                logger.warning(
                    "Runner %s wants Sniper runtime but it is not available; falling back to running without it.",
                    self.name,
                )

        return command

    def get_env(self, os_env=False, disable_runtime=False):
        """Return environment variables used for a game."""
        env = {}
        if os_env:
            env = system.get_environment()

        # Steam compatibility
        if os.environ.get("SteamAppId"):
            logger.info("Game launched from steam (AppId: %s)", os.environ["SteamAppId"])
            env["LC_ALL"] = ""

        # Set correct LC_ALL depending on user settings
        locale = self.system_config.get("locale")
        if locale:
            env["LC_ALL"] = locale

        # By default, we'll set NVidia's shader disk cache to be
        # per-game, so it overflows less readily.
        env["__GL_SHADER_DISK_CACHE"] = "1"
        env["__GL_SHADER_DISK_CACHE_PATH"] = self.nvidia_shader_cache_path

        # Override SDL2 controller configuration
        sdl_gamecontrollerconfig = self.system_config.get("sdl_gamecontrollerconfig")
        if sdl_gamecontrollerconfig:
            path = os.path.expanduser(sdl_gamecontrollerconfig)
            if system.path_exists(path):
                with open(path, "r", encoding="utf-8") as controllerdb_file:
                    sdl_gamecontrollerconfig = controllerdb_file.read()
            env["SDL_GAMECONTROLLERCONFIG"] = sdl_gamecontrollerconfig

        # Set monitor to use for SDL 1 games
        sdl_video_fullscreen = self.system_config.get("sdl_video_fullscreen")
        if sdl_video_fullscreen and sdl_video_fullscreen != "off":
            env["SDL_VIDEO_FULLSCREEN_DISPLAY"] = sdl_video_fullscreen

        if len(GPUS) > 1 and self.system_config.get("gpu") in GPUS:
            gpu = GPUS[self.system_config["gpu"]]
            if gpu.driver == "nvidia":
                env["DRI_PRIME"] = "1"
                env["__NV_PRIME_RENDER_OFFLOAD"] = "1"
                env["__GLX_VENDOR_LIBRARY_NAME"] = "nvidia"
                env["__VK_LAYER_NV_optimus"] = "NVIDIA_only"
            else:
                env["DRI_PRIME"] = gpu.pci_id
            env["VK_ICD_FILENAMES"] = gpu.icd_files  # Deprecated
            env["VK_DRIVER_FILES"] = gpu.icd_files  # Current form

            # To classify for multiple GPUs with the same vendorID:deviceID
            if hasattr(gpu, "device_uuid"):
                env["DXVK_FILTER_DEVICE_UUID"] = gpu.device_uuid

        # Set PulseAudio latency to 60ms
        if self.system_config.get("pulse_latency"):
            env["PULSE_LATENCY_MSEC"] = "60"

        runtime_ld_library_path = None

        if not disable_runtime and self.use_runtime():
            runtime_env = self.get_runtime_env()
            runtime_ld_library_path = runtime_env.get("LD_LIBRARY_PATH")

        if runtime_ld_library_path:
            ld_library_path = env.get("LD_LIBRARY_PATH")
            env["LD_LIBRARY_PATH"] = os.pathsep.join(filter(None, [runtime_ld_library_path, ld_library_path]))

        # Apply user overrides at the end
        env.update(self.system_config.get("env") or {})

        return env

    def finish_env(self, env: Dict[str, str], game) -> None:
        """This is called by the Game after setting up the environment to allow the runner
        to make final adjustments, which may be based on the environment so far."""
        pass

    def get_runtime_env(self):
        """Return runtime environment variables.

        This method may be overridden in runner classes.
        (Notably for Lutris wine builds)

        Returns:
            dict

        """
        return runtime.get_env(prefer_system_libs=self.system_config.get("prefer_system_libs", True))

    def apply_launch_config(self, gameplay_info, launch_config):
        """Updates the gameplay_info to reflect a launch_config section. Called only
        if a non-default config is chosen."""
        gameplay_info["command"] = self.get_launch_config_command(gameplay_info, launch_config)

        config_working_dir = self.get_launch_config_working_dir(launch_config)

        if config_working_dir:
            gameplay_info["working_dir"] = config_working_dir

    def get_launch_config_command(self, gameplay_info, launch_config):
        """Generates a new command for the gameplay_info, to implement the launch_config.
        Returns a new list of strings; the caller can modify it further.

        If launch_config has no command, this builds one from the gameplay_info command
        and the 'exe' value in the launch_config.

        Runners override this when required to control the command used."""

        if "command" in launch_config:
            command = strings.split_arguments(launch_config["command"])
        else:
            command = self.get_command()

        exe = self.get_launch_config_exe(launch_config)
        if exe:
            command.append(exe)

        if launch_config.get("args"):
            command += strings.split_arguments(launch_config["args"])

        return command

    def get_launch_config_exe(self, launch_config):
        """Locates the "exe" of the launch config. If it appears
        to be relative to the game's working_dir, this will try to
        adjust it to be relative to the config's instead.
        """
        exe = launch_config.get("exe")
        config_working_dir = self.get_launch_config_working_dir(launch_config)

        if exe:
            exe = os.path.expanduser(exe)  # just in case!

            if config_working_dir and not os.path.isabs(exe):
                exe_from_config = self.resolve_config_path(exe, config_working_dir)
                exe_from_game = self.resolve_config_path(exe)

                if os.path.exists(exe_from_game) and not os.path.exists(exe_from_config):
                    relative = os.path.relpath(exe_from_game, start=config_working_dir)
                    if not relative.startswith("../"):
                        return relative

        return exe

    def get_launch_config_working_dir(self, launch_config):
        """Extracts the "working_dir" from the config, and resolves this relative
        to the game's working directory, so that an absolute path results.

        This returns None if no working_dir is present, or if it found to be missing.
        """
        config_working_dir = launch_config.get("working_dir")
        if config_working_dir:
            config_working_dir = self.resolve_config_path(config_working_dir)
            if os.path.isdir(config_working_dir):
                return config_working_dir

        return None

    def resolve_config_path(self, path, relative_to=None):
        """Interpret a path taken from the launch_config relative to
        a working directory, using the game's working_dir if that is omitted,
        and expanding the '~' if we get one.

        This is provided as a method so the WINE runner can try to convert
        Windows-style paths to usable paths.
        """
        path = os.path.expanduser(path)

        if not os.path.isabs(path):
            if not relative_to:
                relative_to = self.working_dir

            if relative_to:
                return os.path.join(relative_to, path)

        return path

    def prelaunch(self):
        """Run actions before running the game, override this method in runners; raise an
        exception if prelaunch fails, and it will be reported to the user, and
        then the game won't start."""
        available_libs = set()
        for lib in set(self.require_libs):
            if lib in LINUX_SYSTEM.shared_libraries:
                if self.arch:
                    if self.arch in [_lib.arch for _lib in LINUX_SYSTEM.shared_libraries[lib]]:
                        available_libs.add(lib)
                else:
                    available_libs.add(lib)
        unavailable_libs = set(self.require_libs) - available_libs
        if unavailable_libs:
            raise UnavailableLibrariesError(unavailable_libs, self.arch)

    def get_version(self, use_default=True):
        return None

    def get_run_data(self):
        """Return dict with command (exe & args list) and env vars (dict).

        Reimplement in derived runner if need be."""
        return {"command": self.get_command(), "env": self.get_env()}

    def run(self, ui_delegate):
        """Run the runner alone."""
        if not self.runnable_alone:
            return
        if not self.is_installed():
            if not self.install_dialog(ui_delegate):
                logger.info("Runner install cancelled")
                return

        command_data = self.get_run_data()
        command = command_data.get("command")
        env = (command_data.get("env") or {}).copy()

        self.prelaunch()

        command_runner = MonitoredCommand(command, runner=self, env=env)
        command_runner.start()

    def use_runtime(self):
        if runtime.RUNTIME_DISABLED:
            logger.info("Runtime disabled by environment")
            return False
        if self.system_config.get("disable_runtime"):
            logger.info("Runtime disabled by system configuration")
            return False
        if self.use_sniper_runtime and get_sniper_run_command():
            logger.info("Using Sniper runtime; old Lutris runtime disabled for %s", self.name)
            return False
        return True

    def filter_game_pids(self, candidate_pids: Iterable[int], game_uuid: str, game_folder: str) -> Set[int]:
        """Checks the pids given and returns a set containing only those that are part of the running game,
        identified by its UUID and directory."""
        folder_pids = set()
        gamescope_pids = set()
        has_gamescope = self.system_config.get("gamescope")

        for pid in candidate_pids:
            proc = Process(pid)
            cmdline = proc.cmdline or ""
            # pressure-vessel: This could potentially pick up PIDs not started by lutris?
            if game_folder in cmdline:
                folder_pids.add(pid)
            # Include gamescope-related processes when gamescope is enabled
            if has_gamescope and (proc.name or "").startswith("gamescope"):
                gamescope_pids.add(pid)

        uuid_pids = set(pid for pid in candidate_pids if Process(pid).environ.get("LUTRIS_GAME_UUID") == game_uuid)

        return (folder_pids & uuid_pids) | gamescope_pids

    def install_dialog(self, ui_delegate):
        """Ask the user if they want to install the runner.

        Return success of runner installation.
        """

        if ui_delegate.show_install_yesno_inquiry(
            question=_("The required runner is not installed.\nDo you wish to install it now?"),
            title=_("Required runner unavailable"),
        ):
            version = self.get_version(use_default=False)
            self.install(ui_delegate, version=version)
            return self.is_installed()
        return False

    def is_installed(self, flatpak_allowed: bool = True) -> bool:
        """Return whether the runner is installed"""
        try:
            # Don't care where the exe is, only if we can find it.
            exe = self.get_executable()
            if system.path_exists(exe):
                return True
        except MisconfigurationError:
            pass  # We can still try flatpak even if 'which' fails us!

        return bool(flatpak_allowed and self.flatpak_id and flatpak.is_app_installed(self.flatpak_id))

    def is_installed_for(self, interpreter):
        """Returns whether the runner is installed. Specific runners can extract additional
        script settings, to determine more precisely what must be installed."""
        return self.is_installed()

    def get_installer_runner_version(self, installer, use_runner_config: bool = True) -> Optional[str]:
        return None

    def adjust_installer_runner_config(self, installer_runner_config: Dict[str, Any]) -> None:
        """This is called during installation to let to run fix up in the runner's section of
        the confliguration before it is saved. This method should modify the dict given."""
        pass

    def get_runner_version(self, version: Optional[str] = None) -> Optional[Dict[str, str]]:
        """Get the appropriate version for a runner, as with get_default_runner_version(),
        but this method allows the runner to apply its configuration."""
        return get_default_runner_version_info(self.name, version)

    def install(self, install_ui_delegate, version: Optional[str] = None, callback=None):
        """Install runner using package management systems."""
        logger.debug(
            "Installing %s (version=%s, callback=%s)",
            self.name,
            version,
            callback,
        )
        opts = {"install_ui_delegate": install_ui_delegate, "callback": callback}
        if self.download_url:
            opts["dest"] = self.directory
            return self.download_and_extract(self.download_url, **opts)
        runner_version_info = self.get_runner_version(version)
        if not runner_version_info:
            raise RunnerInstallationError(_("Failed to retrieve {} ({}) information").format(self.name, version))

        if "url" not in runner_version_info:
            if version:
                raise RunnerInstallationError(
                    _("The '%s' version of the '%s' runner can't be downloaded." % (version, self.name))
                )
            else:
                raise RunnerInstallationError(_("The the '%s' runner can't be downloaded." % self.name))

        if "wine" in self.name:
            opts["merge_single"] = True
            opts["dest"] = os.path.join(self.directory, format_runner_version(runner_version_info))

        if self.name == "supermodel":
            opts["merge_single"] = True
            opts["dest"] = os.path.join(settings.RUNNER_DIR, "supermodel")

        if self.name == "libretro" and version:
            opts["merge_single"] = False
            opts["dest"] = os.path.join(settings.RUNNER_DIR, "retroarch/cores")

        self.download_and_extract(runner_version_info["url"], **opts)

    def download_and_extract(self, url, dest=None, **opts):
        install_ui_delegate = opts["install_ui_delegate"]
        merge_single = opts.get("merge_single", False)
        callback = opts.get("callback")
        tarball_filename = os.path.basename(url)
        runner_archive = os.path.join(settings.CACHE_DIR, tarball_filename)
        if not dest:
            dest = settings.RUNNER_DIR

        download_successful = install_ui_delegate.download_install_file(url, runner_archive)
        if download_successful:
            self.extract(archive=runner_archive, dest=dest, merge_single=merge_single, callback=callback)
        else:
            logger.info("Download canceled by the user.")

    def extract(self, archive: str, dest: str, merge_single: bool = False, callback=None):
        if not system.path_exists(archive, exclude_empty=True):
            raise RunnerInstallationError(_("Failed to extract {}").format(archive))
        try:
            extract_archive(archive, dest, merge_single=merge_single)
        except ExtractError as ex:
            logger.error("Failed to extract the archive %s file may be corrupt", archive)
            raise RunnerInstallationError(_("Failed to extract {}: {}").format(archive, ex)) from ex
        os.remove(archive)

        if self.name == "wine":
            logger.debug("Clearing wine version cache")
            from lutris.util.wine.wine import clear_wine_version_cache

            clear_wine_version_cache()

        if self.runner_executable:
            runner_executable = os.path.join(settings.RUNNER_DIR, self.runner_executable)
            if os.path.isfile(runner_executable):
                system.make_executable(runner_executable)

        if callback:
            callback()

    def remove_game_data(self, app_id=None, game_path: Optional[str] = None):
        if game_path:
            system.remove_folder(game_path)

    def can_uninstall(self):
        return os.path.isdir(self.directory)

    def uninstall(self, uninstall_callback: Callable[[], None]) -> None:
        runner_path = self.directory
        if os.path.isdir(runner_path):
            system.remove_folder(runner_path, completion_function=uninstall_callback)
        else:
            uninstall_callback()

    def find_option(self, options_group, option_name):
        """Retrieve an option dict if it exists in the group"""
        if options_group not in ["game_options", "runner_options"]:
            return None
        output = None
        for item in getattr(self, options_group):
            if item["option"] == option_name:
                output = item
                break
        return output

    def force_stop_game(self, game_pids: Iterable[int]) -> None:
        """Stop the running game. If this leaves any game processes running,
        the caller will SIGKILL them (after a delay)."""
        kill_processes(signal.SIGTERM, game_pids)

    def extract_icon(self, game_slug):
        """The config UI calls this to extract the game icon. Most runners do not
        support this and do nothing. This is not called if a custom icon is installed
        for the game."""



"""Steam service"""

import json
import os
from collections import defaultdict
from gettext import gettext as _

from lutris import settings
from lutris.config import LutrisConfig, write_game_config
from lutris.database import sql
from lutris.database.games import add_game, get_game_by_field, get_games
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.installer.installer_file import InstallerFile
from lutris.services.base import BaseService
from lutris.services.lutris import sync_media
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util.log import logger
from lutris.util.steam.appmanifest import AppManifest, get_appmanifests
from lutris.util.steam.config import get_active_steamid64, get_steam_library, get_steamapps_dirs
from lutris.util.strings import slugify


class SteamBanner(ServiceMedia):
    service = "steam"
    size = (184, 69)
    dest_path = os.path.join(settings.CACHE_DIR, "steam/banners")
    file_patterns = ["%s.jpg"]
    api_field = "appid"
    url_pattern = "http://cdn.akamai.steamstatic.com/steam/apps/%s/capsule_184x69.jpg"


class SteamCover(ServiceMedia):
    service = "steam"
    size = (200, 300)
    dest_path = os.path.join(settings.CACHE_DIR, "steam/covers")
    file_patterns = ["%s.jpg"]
    api_field = "appid"
    url_pattern = "http://cdn.steamstatic.com/steam/apps/%s/library_600x900.jpg"


class SteamBannerLarge(ServiceMedia):
    service = "steam"
    size = (460, 215)
    dest_path = os.path.join(settings.CACHE_DIR, "steam/header")
    file_patterns = ["%s.jpg"]
    api_field = "appid"
    url_pattern = "https://cdn.cloudflare.steamstatic.com/steam/apps/%s/header.jpg"


class SteamGame(ServiceGame):
    """ServiceGame for Steam games"""

    service = "steam"
    installer_slug = "steam"
    runner = "steam"

    @classmethod
    def new_from_steam_game(cls, steam_game):
        """Return a Steam game instance from an AppManifest"""
        game = cls()
        game.appid = steam_game["appid"]
        game.game_id = steam_game["appid"]
        game.name = steam_game["name"]
        game.slug = slugify(steam_game["name"])
        game.runner = cls.runner
        game.details = json.dumps(steam_game)
        return game


class SteamService(BaseService):
    id = "steam"
    name = _("Steam")
    icon = "steam-client"
    medias = {
        "banner": SteamBanner,
        "banner_large": SteamBannerLarge,
        "cover": SteamCover,
    }
    default_format = "banner"
    runner = "steam"
    excluded_appids = [
        "221410",  # Steam for Linux
        "228980",  # Steamworks Common Redistributables
        "1070560",  # Steam Linux Runtime
    ]
    game_class = SteamGame

    def load(self):
        """Return importable Steam games"""
        steamid = get_active_steamid64()
        if not steamid:
            logger.error("Unable to find SteamID from Steam config")
            return
        steam_games = get_steam_library(steamid)
        if not steam_games:
            raise RuntimeError(_("Failed to load games. Check that your profile is set to public during the sync."))
        for steam_game in steam_games:
            if steam_game["appid"] in self.excluded_appids:
                continue
            game = self.game_class.new_from_steam_game(steam_game)
            game.save()
        self.match_games()
        return steam_games

    def match_game(self, service_game, lutris_game):
        super().match_game(service_game, lutris_game)

        if service_game:
            # Copy playtimes from Steam's data
            for game in get_games(filters={"service": self.id, "service_id": service_game["appid"]}):
                steam_game_playtime = json.loads(service_game["details"]).get("playtime_forever")
                playtime = steam_game_playtime / 60
                sql.db_update(settings.DB_PATH, "games", {"playtime": playtime}, conditions={"id": game["id"]})

    def get_installer_files(self, installer, _installer_file_id, _selected_extras):
        steam_uri = "$STEAM:%s:."
        appid = str(installer.script["game"]["appid"])
        file = InstallerFile(installer.game_slug, "steam_game", {"url": steam_uri % appid, "filename": appid})
        return [file], []

    def install_from_steam(self, manifest):
        """Create a new Lutris game based on an existing Steam install"""
        if not manifest.is_installed():
            return
        appid = manifest.steamid
        if appid in self.excluded_appids:
            return
        try:
            service_game = ServiceGameCollection.get_game(self.id, appid)
            if not service_game:
                return
            lutris_game_id = "%s-%s" % (self.id, appid)
            existing_game = get_game_by_field(lutris_game_id, "installer_slug")
            if existing_game:
                return
            game_config = LutrisConfig().game_level
            game_config["game"]["appid"] = appid
            configpath = write_game_config(lutris_game_id, game_config)
            slug = self.get_installed_slug(service_game)
            add_game(
                name=service_game["name"],
                runner="steam",
                slug=slug,
                installed=1,
                installer_slug=lutris_game_id,
                configpath=configpath,
                platform="Linux",
                service=self.id,
                service_id=appid,
            )
            return slug
        except Exception as ex:
            logger.error("Failed to install from Steam: %s", ex)
            return None

    @property
    def steamapps_paths(self):
        return get_steamapps_dirs()

    def add_installed_games(self):
        """Syncs installed Steam games with Lutris"""
        stats = {"installed": 0, "removed": 0, "deduped": 0, "paths": []}
        installed_slugs = []
        installed_appids = []

        for steamapps_path in self.steamapps_paths:
            for appmanifest_file in get_appmanifests(steamapps_path):
                if steamapps_path not in stats["paths"]:
                    stats["paths"].append(steamapps_path)
                app_manifest_path = os.path.join(steamapps_path, appmanifest_file)
                try:
                    app_manifest = AppManifest(app_manifest_path)
                    installed_appids.append(app_manifest.steamid)
                    slug = self.install_from_steam(app_manifest)
                    if slug:
                        installed_slugs.append(slug)
                    stats["installed"] += 1
                except Exception as ex:
                    logger.error("Failed to process app manifest %s: %s", app_manifest_path, ex)

        if stats["paths"]:
            logger.debug("%s Steam games detected and installed", stats["installed"])
            logger.debug("Games found in: %s", ", ".join(stats["paths"]))
        else:
            logger.debug("No Steam folder found with games")

        db_games = get_games(filters={"runner": "steam"})
        for db_game in db_games:
            steam_game = Game(db_game["id"])
            if steam_game.config is None:
                logger.warning("Steam game %s has no config", db_game["id"])
                continue
            try:
                appid = steam_game.config.game_level["game"]["appid"]
            except KeyError:
                logger.warning("Steam game %s has no AppID", db_game["id"])
                continue
            if appid not in installed_appids:
                try:
                    steam_game.uninstall()
                    stats["removed"] += 1
                except Exception as ex:
                    logger.error("Failed to uninstall game %s: %s", appid, ex)

        logger.debug("%s Steam games removed", stats["removed"])

        db_appids = defaultdict(list)
        db_games = get_games(filters={"service": "steam"})
        for db_game in db_games:
            db_appids[db_game["service_id"]].append(db_game["id"])

        for _appid, game_ids in db_appids.items():
            if len(game_ids) == 1:
                continue
            for game_id in game_ids:
                steam_game = Game(game_id)
                if steam_game.config is None:
                    logger.warning("Steam game %s has no config for deduplication", game_id)
                    continue
                if not steam_game.playtime:
                    try:
                        # Unsafe to emit a signal from a worker thread!
                        steam_game.uninstall()
                        steam_game.delete()
                        stats["deduped"] += 1
                    except Exception as ex:
                        logger.error("Failed to deduplicate game %s: %s", game_id, ex)

        sync_media(installed_slugs)
        logger.debug("%s Steam games deduplicated", stats["deduped"])

    def generate_installer(self, db_game):
        """Generate a basic Steam installer"""
        return {
            "name": db_game["name"],
            "version": self.name,
            "slug": slugify(db_game["name"]) + "-" + self.id,
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "appid": db_game["appid"],
            "script": {"game": {"appid": db_game["appid"]}},
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def install(self, db_game):
        appid = db_game["appid"]
        db_games = get_games(filters={"service_id": appid, "installed": "1", "service": self.id})
        existing_game = self.match_existing_game(db_games, appid)
        if existing_game:
            logger.debug("Found steam game: %s", existing_game)
            game = Game(existing_game.id)
            game.save()
            return
        self.install_from_api(db_game, appid)



from lutris.util.amazon.protobuf_decoder import (
    Message,
    type_bool,
    type_bytes,
    type_enum,
    type_int64,
    type_string,
    type_uint32,
)


class CompressionAlgorithm:
    none = 0
    lzma = 1


class HashAlgorithm:
    sha256 = 0
    shake128 = 1

    @staticmethod
    def get_name(algorithm):
        if algorithm == HashAlgorithm.sha256:
            return "SHA256"

        if algorithm == HashAlgorithm.shake128:
            return "SHAKE128"

        return None


class SignatureAlgorithm:
    sha256_with_rsa = 0


class CompressionSettings(Message):
    algorithm = CompressionAlgorithm.none

    def __init__(self):
        self.__lookup__ = [("optional", type_enum, "algorithm", 1)]


class Dir(Message):
    path = None
    mode = None

    def __init__(self):
        self.__lookup__ = [("optional", type_string, "path", 1), ("optional", type_uint32, "mode", 2)]


class File(Message):
    path = None
    mode = None
    size = None
    created = None
    hash = None
    hidden = None
    system = None

    def __init__(self):
        self.__lookup__ = [
            ("optional", type_string, "path", 1),
            ("optional", type_uint32, "mode", 2),
            ("optional", type_int64, "size", 3),
            ("optional", type_string, "created", 4),
            ("optional", Hash, "hash", 5),
            ("optional", type_bool, "hidden", 6),
            ("optional", type_bool, "system", 7),
        ]


class Hash(Message):
    algorithm = HashAlgorithm.sha256
    value = None

    def __init__(self):
        self.__lookup__ = [("optional", type_enum, "algorithm", 1), ("optional", type_bytes, "value", 2)]


class Manifest(Message):
    packages = None

    def __init__(self):
        self.__lookup__ = [("repeated", Package, "packages", 1)]


class ManifestHeader(Message):
    compression = None
    hash = None
    signature = None

    def __init__(self):
        self.__lookup__ = [
            ("optional", CompressionSettings, "compression", 1),
            ("optional", Hash, "hash", 2),
            ("optional", Signature, "signature", 3),
        ]


class Package(Message):
    name = None
    files = None
    dirs = None

    def __init__(self):
        self.__lookup__ = [
            ("optional", type_string, "name", 1),
            ("repeated", File, "files", 2),
            ("repeated", Dir, "dirs", 3),
        ]


class Signature(Message):
    algorithm = SignatureAlgorithm.sha256_with_rsa
    value = None

    def __init__(self):
        self.__lookup__ = [("optional", type_enum, "algorithm", 1), ("optional", type_bytes, "value", 2)]
